class StopWatch:
    """
    A stopwatch to measure time intervals. Similar to the stopwatch feature on your phone.
    """

    def __init__(self):
        ...

    def time(self) -> int:
        """
        Gets the current time of the stopwatch.

        Returns:
            Elapsed time in milliseconds.
        """
        return 0

    def pause(self):
        """
        Pauses the stopwatch.
        """
        ...

    def resume(self):
        """
        Resumes the stopwatch.
        """
        ...

    def reset(self):
        """
        Reset the stopwatch time to 0. 

        Note:
            The run state is unaffected:
                - If it was paused, it stays paused (but now at 0).
                - If it was running, it stays running (but starting again from 0)
        """
        ...



from pybricks.parameters import Color


class Light:
    """
    A stub class to represent the light member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def on(self, color: Color):
        """
        Turns on the light at the specified color.

        Args:
            color (Color): Color of the light. The light turns off if you choose None or a color that is not available. The brick status light supports Color.RED, Color.ORANGE, and Color.GREEN.
        """
        ...

    def off(self):
        """
        Turns off the light.
        """
        ...



__all__ = [
    "ColorSensor",
    "EnergyMeter",
    "LightSensor",
    "SoundSensor",
    "TemperatureSensor",
    "TouchSensor",
    "UltrasonicSensor",
    "VernierAdapter"
]

from .__stub.__colorsensor import ColorSensor
from .__stub.__energymeter import EnergyMeter
from .__stub.__lightsensor import LightSensor
from .__stub.__soundsensor import SoundSensor
from .__stub.__temperaturesensor import TemperatureSensor
from .__stub.__touchsensor import TouchSensor
from .__stub.__ultrasonicsensor import UltrasonicSensor
from .__stub.__vernieradapter import VernierAdapter



"""
Module: '_api.speaker' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any


class Speaker():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def beep(self, *args, **kwargs) -> Any:
        """"""

    def get_volume(self, *args, **kwargs) -> Any:
        """"""

    def play_sound(self, *args, **kwargs) -> Any:
        """"""

    def set_volume(self, *args, **kwargs) -> Any:
        """"""

    def start_beep(self, *args, **kwargs) -> Any:
        """"""

    def start_sound(self, *args, **kwargs) -> Any:
        """"""

    def stop(self, *args, **kwargs) -> Any:
        """"""


class SpeakerSoundProvider():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def get_canonical_name(self, *args, **kwargs) -> Any:
        """"""

    def get_sound_files(self, *args, **kwargs) -> Any:
        """"""

system : Any ## <class 'System'> = <System object at 20028790>
def wait_for_async(*args, **kwargs) -> Any:
    """"""




import os
import re
import tkinter as tk
import sys
from tkinter import messagebox, ttk
from PIL import Image, ImageTk


import mysql.connector.pooling

user_table = "sys_user"
# 创建MySQL连接池
config = {
    "user": "root",
    "password": "aY)I&iBop[F3t",
    "host": "180.76.235.69",
    "port": "3306",
    "database": "pyai",
    "auth_plugin": "mysql_native_password",
}

db_pool = mysql.connector.pooling.MySQLConnectionPool(pool_name="mypool", pool_size=5, **config)


# 开启连接
def open_connection():
    # 创建数据库连接
    db_connection = db_pool.get_connection()
    return db_connection


# 关闭连接
def close_connection(cursor, db_connection):
    # 关闭游标和连接
    cursor.close()
    db_connection.close()


class UserLogin:
    def __init__(self):
        self.account = None
        self.password = None
        self.login_image = None
        self.root = tk.Tk()
        self.login_result = False
        self.canvas = None
        self.bg_img = None
        self.account_var = None
        self.account_entry = None

        # 添加"密码："标签和输入框
        self.password_var = None
        self.password_entry = None
        self.root.bind("<KeyPress-Return>", lambda event: self.login())  # 绑定回车键

        self.root.protocol("WM_DELETE_WINDOW", self.null_def)

    # def login(self):
    #     userService = UserService()
    #     account = self.account_var.get()
    #     password = self.password_var.get()
    #
    #     # 校验账号和密码
    #     if not account:
    #         messagebox.showerror("登录结果", "账号不能为空")
    #         return
    #         # 校验手机号码
    #     pattern = r'^1[3457896]\d{9}$'
    #     if not re.match(pattern, account):
    #         messagebox.showerror(title='This is the title', message="电话号码错误", detail='请输入有效的电话号码',
    #                              icon=messagebox.ERROR)
    #         # for child in self.fr.winfo_children():
    #         #     child.configure(state="normal")
    #         self.account_entry.config(state="normal")  # 设置账号输入框为可编辑状态
    #         self.password_entry.config(state="normal")  # 设置密码输入框为可编辑状态
    #
    #         return
    #
    #     if not password:
    #         messagebox.showerror("登录结果", "密码不能为空")
    #         return
    #
    #     if len(password) < 6:
    #         messagebox.showerror("登录结果", "密码不能少于六位")
    #         return
    #     # 执行登录逻辑
    #     if userService.login(account, password):
    #         self.login_result = True
    #         self.on_closing()
    #     else:
    #         # self.config(state="normal")
    #         messagebox.showerror(title="Show Error", message="账号密码错误")

    def login(self):
        db_connection = open_connection()
        cursor = db_connection.cursor(dictionary=True)
        account = self.account_var.get()
        password = self.password_var.get()

        # 校验账号和密码
        if not account:
            messagebox.showerror("登录结果", "账号为身份证后6位，不能为空,")
            return
            # 校验手机号码

        if not password:
            messagebox.showerror("登录结果", "密码为手机号后4位，不能为空")
            return

        if len(password) != 4:
            messagebox.showerror("登录结果", "密码为手机号后4位")
            return

        # query_sql = f"SELECT COUNT(*) FROM '{user_table}' WHERE account = '{account}' AND password = '{password}'"
        query_sql = f"SELECT COUNT(*) FROM {user_table} WHERE SUBSTRING(id_card_num, LENGTH(id_card_num)-5) = {account} AND SUBSTRING(user_name, LENGTH(user_name)-3) ={password}"
        cursor.execute(query_sql)
        fetchone = cursor.fetchone()
        print(fetchone)
        existUser = fetchone['COUNT(*)']
        if existUser > 0:
            query_sql = f"SELECT COUNT(*) FROM {user_table} WHERE SUBSTRING(id_card_num, LENGTH(id_card_num)-5) = {account} AND SUBSTRING(user_name, LENGTH(user_name)-3) = {password} AND state = 1"
            cursor.execute(query_sql)
            fetchone = cursor.fetchone()
            existUser = fetchone['COUNT(*)']
            if existUser > 0:
                self.login_result = True
                self.on_closing()
            else:
                messagebox.showerror(title="Show Error",
                                     message="账号未审核,请先联系管理员审核")

        else:
            query_sql = f"SELECT COUNT(*) FROM {user_table} WHERE SUBSTRING(id_card_num, LENGTH(id_card_num)-5) = {account}"
            cursor.execute(query_sql)
            fetchone = cursor.fetchone()
            print(fetchone)
            existUser = fetchone['COUNT(*)']
            if existUser > 0:
                messagebox.showerror(title="Show Error", message="密码错误,密码为手机号后4位")
            else:
                messagebox.showerror(title="Show Error",
                                     message="账号错误,账号为身份证后6位,若无账户请先前往公众号注册")

        close_connection(cursor, db_connection)

    def get_current_os(self):
        platform = sys.platform
        if platform.startswith('win'):
            return "Windows"
        elif platform.startswith('linux'):
            return "Linux"
        elif platform == 'darwin':
            return "Mac"
        else:
            return "Unknown"

    def create_login_view(self):
        root = self.root
        root.title("引航计划")
        # root.resizable(width=False, height=False)
        # root.overrideredirect(True)
        root.geometry("{0}x{1}+0+0".format(root.winfo_screenwidth(), root.winfo_screenheight()))

        self.canvas = tk.Canvas(root)
        self.canvas.pack(fill="both", expand=True)
        system_type = self.get_current_os()
        if system_type == 'Windows':
            root.wm_attributes("-toolwindow", True)
        elif system_type == 'Linux':
            root.protocol("WM_ICONIFY", self.null_def)
            root.bind("<Unmap>", self.null_def)
        login_bg = os.path.join(os.path.dirname(__file__), "res", "login-poster.png")
        self.bg_img = Image.open(login_bg)
        self.bg_img = self.bg_img.resize((root.winfo_screenwidth(), root.winfo_screenheight()), Image.LANCZOS)
        self.bg_img = ImageTk.PhotoImage(self.bg_img)

        # Create a background image item on the canvas and set its anchor to 'nw'
        self.canvas.create_image(0, 0, image=self.bg_img, anchor="nw", tags="bg")
        self.canvas.config(highlightthickness=0)

        # 获取 Canvas 的宽度和高度
        canvas_width = root.winfo_screenwidth()
        canvas_height = root.winfo_screenheight()

        img_dir = os.path.join(os.path.dirname(__file__), "res", "login_logo.png")
        original_image = Image.open(img_dir)
        resized_image = original_image.resize((320, 130), Image.LANCZOS)  # Resize to 100x100
        self.login_image = ImageTk.PhotoImage(resized_image)
        # 获取图像的宽度和高度
        image_width = self.login_image.width()
        image_height = self.login_image.height()

        # 计算图像在 Canvas 上居中的位置
        x = (canvas_width - image_width) / 2
        y = (canvas_height - image_height) / 2
        self.canvas.create_image(x, y - 150, image=self.login_image, anchor="nw")
        # 添加"账号："标签和输入框
        self.canvas.create_text(x + 40, y + 15, text="账号：".encode("utf-8").decode("utf-8"), anchor="nw")
        self.account_var = tk.StringVar()
        self.account_entry = ttk.Entry(self.canvas, textvariable=self.account_var)
        self.canvas.create_window(x + 190, y + 28, window=self.account_entry)
        # self.account_entry.place(relx=0.46, rely=0.452, relwidth=0.11, relheight=0.025)

        # 添加"密码："标签和输入框
        self.canvas.create_text(x + 40, y + 65, text="密码：".encode("utf-8").decode("utf-8"), anchor="nw")
        self.password_var = tk.StringVar()
        self.password_entry = ttk.Entry(self.canvas, show="*", textvariable=self.password_var)
        self.canvas.create_window(x + 190, y + 75, window=self.password_entry)

        button = ttk.Button(self.canvas, text="登录".encode("utf-8").decode("utf-8"), command=self.login)
        self.account_entry.configure(state="normal")
        self.password_entry.configure(state="normal")
        self.canvas.create_window(x + 160, y + 140, window=button)

        root.mainloop()

    def on_closing(self):
        self.root.destroy()  # 销毁窗口

    def null_def(self):
        pass



"""
MIT License

Copyright (c) 2023 Aivar Annamaa

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import subprocess
import sys
import textwrap
from typing import List, Optional

from pipkin import main

logger = logging.getLogger(__name__)

MP_ORG_INDEX = "https://micropython.org/pi"
PYPI_INDEX = "https://pypi.org/pypi"
PYPI_SIMPLE_INDEX = "https://pypi.org/simple"
DEFAULT_INDEX_URLS = [MP_ORG_INDEX, PYPI_INDEX]
SERVER_ENCODING = "utf-8"

__version__ = "0.2b1"

"""
steps:
    - infer target if no explicit connection parameters are given
    - connect (MP)
    - determine target location on the device/mount
    - sync RTC (MP, install, uninstall). Not required for CP?
    - ensure temp venv for pip operations
    - fetch METADATA-s and RECORD-s (may be empty in all cases except "show -f") and populate venv
    - record current state
    - invoke pip (translate paths in the output)
    - determine deleted and changed dists and remove these on the target (according to actual RECORD-s)
    - determine new and changed dists and copy these to the target
    - clear venv



"""


sys.exit(main(sys.argv[1:]))



"""Touch related IO

The `touchio` module contains classes to provide access to touch IO typically
accelerated by hardware on the onboard microcontroller.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For more information about working with the `touchio` module in CircuitPython,
see `this Learn guide page <https://learn.adafruit.com/circuitpython-essentials/circuitpython-cap-touch>`_.

Example::

  import touchio
  from board import *

  touch_pin = touchio.TouchIn(D6)
  print(touch_pin.value)

This example will initialize the the device, and print the
:py:data:`~touchio.TouchIn.value`."""

from __future__ import annotations

from typing import Optional

import microcontroller

class TouchIn:
    """Read the state of a capacitive touch sensor

    Usage::

       import touchio
       from board import *

       touch = touchio.TouchIn(A1)
       while True:
           if touch.value:
               print("touched!")"""

    def __init__(self, pin: microcontroller.Pin) -> None:
        """Use the TouchIn on the given pin.

        :param ~microcontroller.Pin pin: the pin to read from"""
        ...
    def deinit(self) -> None:
        """Deinitialises the TouchIn and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> TouchIn:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    value: bool
    """Whether the touch pad is being touched or not. (read-only)

    True when `raw_value` > `threshold`."""
    raw_value: int
    """The raw touch measurement as an `int`. (read-only)"""
    threshold: Optional[int]
    """Minimum `raw_value` needed to detect a touch (and for `value` to be `True`).

    When the **TouchIn** object is created, an initial `raw_value` is read from the pin,
    and then `threshold` is set to be 100 + that value.

    You can adjust `threshold` to make the pin more or less sensitive::

      import board
      import touchio

      touch = touchio.TouchIn(board.A1)
      touch.threshold = 7300"""



from __future__ import annotations

import sys
import tkinter as tk
from logging import getLogger
from typing import Dict, List, Literal, Optional, Tuple, Union, cast

from turcar import dnd  # inspired by and very similar to tkinter.dnd
from turcar.languages import tr

logger = getLogger(__name__)


class CustomNotebookTabRow(tk.Frame):
    def __init__(self, master: CustomNotebook, **kw):
        self.notebook = master
        self.root = master.winfo_toplevel()
        super().__init__(master, **kw)
        self.filler = tk.Frame(self, background=self.notebook.base_style_conf["background"])
        self.filler.grid(row=0, column=999, sticky="nsew", padx=(1, 0), pady=(0, 1))
        self._insertion_mark = tk.Frame(
            self, background="black", width=_ems_to_pixels(0.3), height=_ems_to_pixels(2)
        )
        self._insertion_index = None

    def dnd_accept(self, source, event):
        assert isinstance(source, CustomNotebookTab)
        return self

    def dnd_enter(self, source, event):
        self.dnd_motion(source, event)

    def dnd_motion(self, source: CustomNotebookTab, event: tk.Event):
        x = event.x_root - self.winfo_rootx()
        y = event.y_root - self.winfo_rooty()
        insertion_index, insertion_widget = self._compute_insertion_position(x, y)
        source_index = source.notebook.index_of_tab(source)
        if source.notebook is self.notebook and insertion_index in [source_index, source_index + 1]:
            # no point in moving a tab right before or after itself
            self._insertion_mark.place_forget()
        else:
            self._show_insertion_mark(insertion_index, insertion_widget)

    def dnd_leave(self, source, event):
        self.hide_insertion_mark()

    def dnd_commit(self, source, event):
        logger.info("dnd_commit with insertion_index %r", self._insertion_index)
        if self._insertion_index is not None:
            self.notebook.insert_from_another_notebook_or_position(
                self._insertion_index, source.notebook.get_page_by_tab(source)
            )
        self.dnd_leave(source, event)

    def _compute_insertion_position(self, mouse_x: int, mouse_y: int) -> Tuple[int, tk.Widget]:
        dist_from_closest_left_edge = None
        closest_left_edge_widget = None
        closest_left_edge_index = None
        items = [page.tab for page in self.notebook.pages] + [self.filler]
        for i, item in enumerate(items):
            dist_from_left_edge = abs(item.winfo_x() - mouse_x)
            if (
                dist_from_closest_left_edge is None
                or dist_from_left_edge < dist_from_closest_left_edge
            ):
                dist_from_closest_left_edge = dist_from_left_edge
                closest_left_edge_widget = item
                closest_left_edge_index = i

        return closest_left_edge_index, closest_left_edge_widget

    def _show_insertion_mark(self, insertion_index: int, right_neigbour: tk.Widget) -> None:
        self._insertion_mark.place_configure(
            x=right_neigbour.winfo_x(), y=0, height=self.winfo_height()
        )
        self._insertion_index = insertion_index
        self._insertion_mark.lift()

    def hide_insertion_mark(self) -> None:
        self._insertion_mark.place_forget()
        self._insertion_index = None


class CustomNotebook(tk.Frame):
    def __init__(self, master: Union[tk.Widget, tk.Toplevel, tk.Tk], closable: bool = True):
        self.base_style_conf = _get_style_configuration(".")
        self.style_conf = _get_style_configuration("CustomNotebook")
        super().__init__(master, background=self.style_conf["bordercolor"])
        self.closable = closable
        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)

        # Can't use ttk.Frame, because can't change its color in aqua
        self.tab_row = CustomNotebookTabRow(self, background=self.style_conf["bordercolor"])
        self.tab_row.grid(row=0, column=0, sticky="new")
        self.tab_row.columnconfigure(999, weight=1)

        self.current_page: Optional[CustomNotebookPage] = None
        self.pages: List[CustomNotebookPage] = []

    def add(self, child: tk.Widget, text: str) -> None:
        self._plain_insert("end", child, text=text)

    def insert(
        self, pos: Union[int, Literal["end"], tk.Widget], child: tk.Widget, text: str
    ) -> None:
        if pos == "auto":
            child_position_key = getattr(child, "position_key", None)
            if child_position_key is not None:
                for sibling_content in [page.content for page in self.pages]:
                    sibling_position_key = getattr(sibling_content, "position_key", None)

                    if sibling_position_key is None or sibling_position_key > child_position_key:
                        pos = sibling_content
                        break
                else:
                    pos = "end"
            else:
                logger.warning("auto insert without position_key")
                pos = "end"

        self._plain_insert(pos, child, text)

    def _plain_insert(
        self, pos: Union[int, Literal["end"], tk.Widget], child: tk.Widget, text: str
    ) -> None:
        tab = CustomNotebookTab(self, title=text, closable=self.closable)
        page = CustomNotebookPage(tab, child)
        if isinstance(pos, tk.Widget):
            pos = self.index(pos)

        if pos == "end":
            self.pages.append(page)
        else:
            logger.info("Inserting %r to position %r in %r", child, pos, self.pages)
            self.pages.insert(pos, page)

        self._rearrange_tabs()
        if len(self.pages) == 1:
            self.select_tab(page.tab)
        child.containing_notebook = self
        self.after_add_or_insert(page)

    def add_from_another_notebook(self, page: CustomNotebookPage) -> None:
        self.insert_from_another_notebook_or_position("end", page)

    def insert_from_another_notebook_or_position(
        self, pos: Union[int, Literal["end"]], page: CustomNotebookPage
    ) -> None:
        original_notebook = page.tab.notebook
        if original_notebook == self and pos >= self.index_of_tab(page.tab):
            # tab will be removed first, so the new target index will be one less
            pos -= 1

        original_notebook._plain_forget(page.content)
        self._plain_insert(pos, page.content, text=page.tab.get_title())
        setattr(page.content, "containing_notebook", self)

    def _rearrange_tabs(self) -> None:
        for i, page in enumerate(self.pages):
            page.tab.grid(row=0, column=i, sticky="nsew", padx=(1, 0), pady=(1, 0))

    def enable_traversal(self) -> None:
        # TODO:
        pass

    def select(self, tab_id: Union[str, tk.Widget, None] = None) -> Optional[str]:
        if tab_id is None:
            if self.current_page:
                return self.current_page.content.winfo_name()
            return None

        return self.select_by_index(self.index(tab_id))

    def select_by_index(self, index: int) -> None:
        new_page = self.pages[index]
        if new_page == self.current_page:
            return

        new_page.content.grid_propagate(False)
        new_page.content.grid(row=1, column=0, sticky="nsew", padx=(1, 1), pady=(0, 1), in_=self)
        if self.current_page and self.current_page.content.winfo_exists():
            self.current_page.content.grid_remove()
            # new_page.content.tkraise(self.current_page.content)
        print("new", new_page.content)
        new_page.tab.update_state(True)
        if self.current_page:
            print("old", self.current_page.content)
            self.current_page.tab.update_state(False)

        print("Current page becomes", new_page.content)
        self.current_page = new_page
        self.event_generate("<<NotebookTabChanged>>")

    def select_tab(self, tab: CustomNotebookTab) -> None:
        for i, page in enumerate(self.pages):
            if page.tab == tab:
                self.select_by_index(i)
                return

        raise ValueError(f"Unknown tab {tab}")

    def select_page(self, page: CustomNotebookPage) -> None:
        self.select_by_index(self.pages.index(page))

    def index(self, tab_id: Union[str, tk.Widget]) -> int:
        if tab_id == "end":
            return len(self.pages)

        for i, page in enumerate(self.pages):
            if page.content == tab_id or page.content.winfo_name() == tab_id:
                return i
        else:
            raise RuntimeError(f"Can't find {tab_id!r}")

    def index_of_tab(self, tab: CustomNotebookTab) -> int:
        for i, page in enumerate(self.pages):
            if page.tab is tab:
                return i
        else:
            raise RuntimeError(f"Can't find {tab!r}")

    def tab(self, tab_id: Union[str, tk.Widget], text: str):
        page = self.pages[self.index(tab_id)]
        page.tab.set_title(text)

    def tabs(self) -> List[str]:
        return [page.content.winfo_name() for page in self.pages]

    def winfo_children(self) -> List[tk.Widget]:
        return [page.content for page in self.pages]

    def forget(self, child: tk.Widget) -> None:
        self._plain_forget(child)

    def _plain_forget(self, child: tk.Widget) -> None:
        for i, page in enumerate(self.pages):
            if child is page.content:
                break
        else:
            raise ValueError(f"Can't find {child}")

        if page == self.current_page:
            # Choose new active page
            if len(self.pages) > i + 1:
                self.select_page(self.pages[i + 1])  # prefer right neighbor
            elif i > 0:
                self.select_page(self.pages[i - 1])  # left neighbor

        child = self.pages[i].content
        child.grid_forget()
        self.pages[i].tab.grid_forget()

        del self.pages[i]
        self._rearrange_tabs()

        child.containing_notebook = None
        self.after_forget(page)

    def after_add_or_insert(self, page: CustomNotebookPage) -> None:
        self.event_generate("<<NotebookTabOpened>>")

    def after_forget(self, page: CustomNotebookPage) -> None:
        self.event_generate("<<NotebookTabClosed>>")

    def get_child_by_index(self, index: int) -> tk.Widget:
        return self.pages[index].content

    def get_current_child(self) -> Optional[tk.Widget]:
        if self.current_page:
            return self.current_page.content
        return None

    def has_content(self, child: tk.Widget) -> bool:
        for page in self.pages:
            if page.content is child:
                return True

        return False

    def focus_set(self):
        if self.current_page:
            self.current_page.content.focus_set()
        else:
            super().focus_set()

    def get_page_by_tab(self, tab: CustomNotebookTab) -> CustomNotebookPage:
        for page in self.pages:
            if page.tab == tab:
                return page

        raise ValueError(f"Could not find tab {tab}")

    def close_tab(self, index_or_tab: Union[int, CustomNotebookTab]) -> None:
        if isinstance(index_or_tab, int):
            page = self.pages[index_or_tab]
        else:
            page = self.get_page_by_tab(index_or_tab)

        self._plain_forget(page.content)

    def close_tabs(self, except_tab: Optional[CustomNotebookTab] = None):
        for page in reversed(self.pages):
            if page.tab == except_tab:
                continue
            else:
                self.close_tab(page.tab)

    def allows_dragging_to_another_notebook(self) -> bool:
        return False


class CustomNotebookTab(tk.Frame):
    close_image = None
    active_close_image = None

    def __init__(self, notebook: CustomNotebook, title: str, closable: bool):
        super().__init__(notebook.tab_row, borderwidth=0)

        self.base_style = _get_style_configuration(".")
        self.tab_style = _get_style_configuration("CustomNotebook.Tab")
        self.notebook_style_conf = _get_style_configuration("CustomNotebook")

        self.notebook = notebook
        self.title = title
        self.drag_start_x_root = 0
        self.drag_start_y_root = 0

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.label = tk.Label(self, text=title, foreground=self.base_style["foreground"])
        self.label.grid(
            row=0,
            column=0,
            padx=(_ems_to_pixels(0.3), _ems_to_pixels(0.1)),
            sticky="nsw",
            pady=(0, _ems_to_pixels(0.1)),
        )
        self.bind("<1>", self.on_click, True)
        self.label.bind("<1>", self.on_click, True)

        if closable:
            if not CustomNotebookTab.close_image:
                CustomNotebookTab.close_image = _get_image("tab-close")
                CustomNotebookTab.active_close_image = _get_image("tab-close-active")
            self.button = tk.Label(self, image=CustomNotebookTab.close_image)
            self.button.grid(row=0, column=1, padx=(0, _ems_to_pixels(0.1)))
            self.button.bind("<1>", self.on_button_click, True)
            self.button.bind("<Enter>", self.on_button_enter, True)
            self.button.bind("<Leave>", self.on_button_leave, True)

            if sys.platform == "darwin":
                self.label.bind("<ButtonPress-2>", self._right_btn_press, True)
                self.label.bind("<Control-Button-1>", self._right_btn_press, True)
                self.label.bind("<ButtonPress-3>", self._middle_btn_press, True)
            else:
                self.label.bind("<ButtonPress-3>", self._right_btn_press, True)
                self.label.bind("<ButtonPress-2>", self._middle_btn_press, True)

        else:
            self.button = None

        self.indicator = tk.Frame(
            self, height=1, background=self.notebook_style_conf["bordercolor"]
        )
        self.indicator.grid(row=1, column=0, columnspan=2, sticky="sew")

        self.menu = tk.Menu(
            self.winfo_toplevel(), tearoff=False, **_get_style_configuration("Menu")
        )
        self.menu.add_command(label=tr("Close"), command=self._close_tab)
        self.menu.add_command(label=tr("Close others"), command=self._close_other_tabs)
        self.menu.add_command(label=tr("Close all"), command=self._close_all_tabs)

    def set_title(self, text: str) -> None:
        self.label.configure(text=text)

    def get_title(self) -> str:
        return self.label.cget("text")

    def _right_btn_press(self, event):
        self.menu.tk_popup(*self.winfo_toplevel().winfo_pointerxy())

    def _middle_btn_press(self, event):
        self._close_tab()

    def _close_tab(self) -> None:
        self.notebook.close_tab(self)

    def _close_all_tabs(self) -> None:
        self.notebook.close_tabs()

    def _close_other_tabs(self) -> None:
        self.notebook.close_tabs(except_tab=self)

    def on_click(self, event: tk.Event):
        self.notebook.select_tab(self)
        self.drag_start_x_root = event.x_root
        self.drag_start_y_root = event.y_root
        dnd.dnd_start(self, event)

    def on_button_click(self, event):
        self.notebook.close_tab(self)

    def on_button_enter(self, event):
        self.button.configure(image=CustomNotebookTab.active_close_image)

    def on_button_leave(self, event):
        self.button.configure(image=CustomNotebookTab.close_image)

    def update_state(self, active: bool) -> None:
        if active:
            main_background = self.tab_style["activebackground"]
            # indicator_background = "systemTextBackgroundColor"
            # indicator_height = 1

            # indicator_background = border_color
            # indicator_height = 1

            indicator_background = self.tab_style["indicatorbackground"]
            indicator_height = self.tab_style.get("indicatorheight", _ems_to_pixels(0.2))
        else:
            main_background = self.tab_style["background"]
            indicator_background = self.notebook_style_conf["bordercolor"]
            indicator_height = 1

        self.configure(background=main_background)
        self.label.configure(background=main_background)
        if self.button:
            self.button.configure(background=main_background)
        self.indicator.configure(background=indicator_background, height=indicator_height)

    def dnd_motion_anywhere(self, source: CustomNotebookTab, event: tk.Event):
        toplevel = self.winfo_toplevel()
        delta_x = abs(source.drag_start_x_root - event.x_root)
        delta_y = abs(source.drag_start_y_root - event.y_root)
        moved_enough = _ems_to_pixels(2)
        if (
            (delta_x > moved_enough or delta_y > moved_enough)
            and hasattr(toplevel, "show_notebook_drop_targets")
            and self.notebook.allows_dragging_to_another_notebook()
        ):
            toplevel.show_notebook_drop_targets()

    def dnd_end(self, target, event):
        logger.info("Ending DND operation with target %r", target)
        toplevel = self.winfo_toplevel()
        if (
            hasattr(toplevel, "hide_notebook_drop_targets")
            and self.notebook.allows_dragging_to_another_notebook()
        ):
            toplevel.hide_notebook_drop_targets()


class CustomNotebookPage:
    def __init__(self, tab: CustomNotebookTab, content: tk.Widget):
        self.tab = tab
        self.content = content


class NotebookTabDropTarget(tk.Canvas):
    def __init__(
        self,
        master: Union[tk.Tk, tk.Toplevel, tk.Widget],
        notebook: CustomNotebook,
        width: int,
        height: int,
    ):
        self.notebook = notebook
        self.inactive_color = "#E4E4E4"
        self.active_color = "gray"
        super().__init__(
            master=master,
            width=width,
            height=height,
            borderwidth=0,
            highlightthickness=0,
            background=self.inactive_color,
        )

        borderwidth = _ems_to_pixels(0.2)
        dash_step = _ems_to_pixels(0.5)
        self.rect_id = self.create_rectangle(
            1,
            1,
            width - borderwidth,
            height - borderwidth,
            dash=(dash_step, dash_step * 2),
            width=borderwidth,
            fill=self.inactive_color,
        )

    def dnd_accept(self, source: CustomNotebookTab, event: tk.Event):
        assert isinstance(source, CustomNotebookTab)
        return self

    def dnd_enter(self, source: CustomNotebookTab, event: tk.Event):
        self.configure(background=self.active_color)
        self.itemconfigure(self.rect_id, fill=self.active_color)
        self.dnd_motion(source, event)

    def dnd_motion(self, source: CustomNotebookTab, event: tk.Event):
        pass

    def dnd_leave(self, source: CustomNotebookTab, event: tk.Event):
        self.configure(background=self.inactive_color)
        self.itemconfigure(self.rect_id, fill=self.inactive_color)

    def dnd_commit(self, source: CustomNotebookTab, event: tk.Event):
        self.dnd_leave(source, event)
        self.notebook.add_from_another_notebook(source.notebook.get_page_by_tab(source))


def _get_image(name: str) -> tk.PhotoImage:
    from turcar import get_workbench

    return get_workbench().get_image(name)


def _ems_to_pixels(x: float) -> int:
    from turcar.ui_utils import ems_to_pixels

    return ems_to_pixels(x)


def _get_style_configuration(name: str) -> Dict:
    from turcar.ui_utils import get_style_configuration

    return get_style_configuration(name)


"""

class TextFrame(tk.Frame):
    def __init__(self, master):
        super().__init__(master)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.text = tk.Text(self, highlightthickness=0, borderwidth=0, width=50, height=20)
        # self.text.grid_propagate(True)
        self.text.grid(row=0, column=0, sticky="nsew")

        self.scrollbar = ttk.Scrollbar(self, orient="vertical")
        self.scrollbar.grid(row=0, column=1, sticky="nsew")

        if sys.platform == "darwin":
            root = self.winfo_toplevel()
            isdark = int(root.eval(f"tk::unsupported::MacWindowStyle isdark {root}"))
            # Not sure if it is good idea to use fixed colors, but no named (light-dark aware) color matches.
            # Best dynamic alternative is probably systemTextBackgroundColor
            if isdark:
                stripe_color = "#2d2e31"
                print("Dark")
            else:
                stripe_color = "#fafafa"
            stripe = tk.Frame(self, width=1, background=stripe_color)
            stripe.grid(row=0, column=1, sticky="nse")
            stripe.tkraise()

        self.scrollbar["command"] = self.text.yview
        self.text["yscrollcommand"] = self.scrollbar.set


if __name__ == "__main__":
    if sys.platform == "win32":
        import ctypes

        PROCESS_SYSTEM_DPI_AWARE = 1
        ctypes.OleDLL("shcore").SetProcessDpiAwareness(PROCESS_SYSTEM_DPI_AWARE)

    root = tk.Tk()
    root.geometry("800x600")

    style = ttk.Style()
    # style.theme_use("aqua")

    nb = CustomNotebook(root)

    for i in range(4):
        tf = TextFrame(nb)
        tf.text.insert("end", "print('hello world')\n" * i * 30)
        nb.add(tf, f"program{i}.py")

    nb.grid(sticky="nsew", row=0, column=0, padx=15, pady=15)
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)

    root.mainloop()
"""



import os.path
import re
import subprocess
import sys
from dataclasses import dataclass
from logging import getLogger
from typing import List, Optional, Set, Tuple

import pkg_resources

logger = getLogger(__name__)


@dataclass
class ParsedWheelFilename:
    project: str
    version: str
    build: Optional[str]
    python_tags: list[str]
    abi_tags: list[str]
    platform_tags: list[str]


def parse_wheel_filename(filename: str) -> ParsedWheelFilename:
    # Adapted from https://github.com/jwodder/wheel-filename/blob/1568eb2f1726425588550067f09f5c0fde6c9652/src/wheel_filename/__init__.py
    PYTHON_TAG_RGX = r"[\w\d]+"
    ABI_TAG_RGX = r"[\w\d]+"
    PLATFORM_TAG_RGX = r"[\w\d]+"

    WHEEL_FILENAME_CRGX = re.compile(
        r"(?P<project>[A-Za-z0-9](?:[A-Za-z0-9._]*[A-Za-z0-9])?)"
        r"-(?P<version>[A-Za-z0-9_.!+]+)"
        r"(?:-(?P<build>[0-9][\w\d.]*))?"
        r"-(?P<python_tags>{0}(?:\.{0})*)"
        r"-(?P<abi_tags>{1}(?:\.{1})*)"
        r"-(?P<platform_tags>{2}(?:\.{2})*)"
        r"\.[Ww][Hh][Ll]".format(PYTHON_TAG_RGX, ABI_TAG_RGX, PLATFORM_TAG_RGX)
    )
    basename = os.path.basename(os.fsdecode(filename))
    m = WHEEL_FILENAME_CRGX.fullmatch(basename)
    if not m:
        raise ValueError(f"Unexpected wheel filename {basename}")

    return ParsedWheelFilename(
        project=m.group("project"),
        version=m.group("version"),
        build=m.group("build"),
        python_tags=m.group("python_tags").split("."),
        abi_tags=m.group("abi_tags").split("."),
        platform_tags=m.group("platform_tags").split("."),
    )


def create_dist_info_version_name(dist_name: str, version: str) -> str:
    # https://packaging.python.org/en/latest/specifications/binary-distribution-format/#escaping-and-unicode
    # https://peps.python.org/pep-0440/
    safe_name = pkg_resources.safe_name(dist_name).replace("-", "_")
    safe_version = pkg_resources.safe_version(version)
    return f"{safe_name}-{safe_version}"


def get_windows_folder(ID: int) -> str:
    # http://stackoverflow.com/a/3859336/261181
    # http://www.installmate.com/support/im9/using/symbols/functions/csidls.htm
    if sys.platform == "win32":
        import ctypes.wintypes

        SHGFP_TYPE_CURRENT = 0
        buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(0, ID, 0, SHGFP_TYPE_CURRENT, buf)
        assert buf.value
        return buf.value
    else:
        raise AssertionError("Meant to be used only on Windows")


def get_windows_roaming_appdata_dir() -> str:
    return get_windows_folder(26)


def get_windows_local_appdata_dir() -> str:
    return get_windows_folder(28)


def get_user_cache_dir() -> str:
    if sys.platform == "win32":
        return os.path.join(get_windows_local_appdata_dir())
    elif sys.platform == "darwin":
        return os.path.expanduser("~/Library/Caches")
    else:
        return os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))


def get_base_executable():
    if sys.exec_prefix == sys.base_exec_prefix:
        return sys.executable

    if sys.platform == "win32":
        guess = sys.base_exec_prefix + "\\" + os.path.basename(sys.executable)
        if os.path.isfile(guess):
            return guess

    if os.path.islink(sys.executable):
        return os.path.realpath(sys.executable)

    raise RuntimeError("Don't know how to locate base executable")


def get_venv_executable(path: str) -> str:
    if sys.platform == "win32":
        return os.path.join(path, "Scripts", "python.exe")
    else:
        return os.path.join(path, "bin", "python3")


def get_venv_site_packages_path(venv_path: str) -> str:
    logger.debug("Querying site packages path for %s", venv_path)
    args = [
        get_venv_executable(venv_path),
        "-c",
        "import site; print([p for p in site.getsitepackages() if 'site-packages' in p][0])",
    ]
    result = subprocess.check_output(
        args,
        executable=args[0],
        text=True,
        stdin=subprocess.DEVNULL,
    ).strip()
    assert result.startswith(venv_path) and result != venv_path
    logger.debug("Got site packages path %s", result)
    return result


def parse_meta_dir_name(name: str) -> Tuple[str, str]:
    assert name.endswith(".dist-info")
    name, version = name[: -len(".dist-info")].split("-")
    return name, version


def parse_dist_file_name(file_name: str) -> Tuple[str, str, str]:
    file_name = file_name.lower()

    if file_name.endswith(".whl"):
        pwf = parse_wheel_filename(file_name)
        return pwf.project, pwf.version, ".whl"

    for suffix in [".zip", ".tar.gz"]:
        if file_name.endswith(suffix):
            file_name = file_name[: -len(suffix)]
            break
    else:
        raise AssertionError("Unexpected file name " + file_name)

    # dist name and version is separated by the dash, but both parts can also contain dashes...
    if file_name.count("-") == 1:
        dist_name, version = file_name.split("-")
    else:
        # assuming dashes in the version part have digit on their left and letter on their right
        # let's get rid of these
        tweaked_file_name = re.sub(r"(\d)-([a-zA-Z])", r"\1_\2", file_name)
        # now let's treat the rightmost dash as separator
        dist_name = tweaked_file_name.rsplit("-", maxsplit=1)[0]
        version = file_name[len(dist_name) + 1 :]

    return dist_name, version, suffix


def starts_with_continuation_byte(data: bytes) -> bool:
    return len(data) > 0 and is_continuation_byte(data[0])


def is_continuation_byte(byte: int) -> bool:
    return (byte & 0b11000000) == 0b10000000


def custom_normalize_dist_name(name: str) -> str:
    # https://peps.python.org/pep-0503/#normalized-names
    return pkg_resources.safe_name(name).lower().replace("-", "_").replace(".", "_")


def list_volumes(skip_letters: Optional[Set[str]] = None) -> List[str]:
    skip_letters = skip_letters or set()

    "Adapted from https://github.com/ntoll/uflash/blob/master/uflash.py"
    if sys.platform == "win32":
        import ctypes

        #
        # In certain circumstances, volumes are allocated to USB
        # storage devices which cause a Windows popup to raise if their
        # volume contains no media. Wrapping the check in SetErrorMode
        # with SEM_FAILCRITICALERRORS (1) prevents this popup.
        #
        old_mode = ctypes.windll.kernel32.SetErrorMode(1)  # @UndefinedVariable
        try:
            volumes = []
            for disk in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                if disk in skip_letters:
                    continue
                path = "{}:\\".format(disk)
                if os.path.exists(path):
                    volumes.append(path)

            return volumes
        finally:
            ctypes.windll.kernel32.SetErrorMode(old_mode)  # @UndefinedVariable
    else:
        # 'posix' means we're on Linux or OSX (Mac).
        # Call the unix "mount" command to list the mounted volumes.
        mount_output = subprocess.check_output(["mount"], stdin=subprocess.DEVNULL).splitlines()
        return [x.split()[2].decode("utf-8") for x in mount_output]



import os.path
import tkinter as tk
import tkinter.font
from tkinter import ttk

import turcar
from turcar import get_workbench, tktextext, ui_utils
from turcar.config import try_load_configuration
from turcar.languages import tr
from turcar.tktextext import TextFrame
from turcar.ui_utils import scrollbar_style


class HelpView(TextFrame):
    def __init__(self, master):
        from turcar import rst_utils

        TextFrame.__init__(
            self,
            master,
            text_class=rst_utils.RstText,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            borderwidth=0,
            wrap="word",
            relief="flat",
            padx=20,
            pady=0,
            read_only=True,
        )
        self.language_dir = os.path.join(
            os.path.dirname(turcar.__file__),
            "locale",
            get_workbench().get_option("general.language"),
            "HELP_CONTENT",
        )
        get_workbench().bind("WorkbenchReady", self.on_workbench_ready, True)

    def on_workbench_ready(self, event=None):
        self.load_index()

    def on_show(self):
        self.load_index()

    def load_index(self):
        self.load_rst_file("index.rst")

    def load_topic(self, topic, fragment=None):
        self.load_rst_file(topic + ".rst")
        # TODO: scroll to fragment

    def load_rst_file(self, filename):
        self.text.clear()
        self.text.direct_insert("1.0", "\n")

        if os.path.isabs(filename):
            full_path = filename
        else:
            # try to access filename in a language subdirectory
            full_path = os.path.join(self.language_dir, filename)
            if not os.path.exists(full_path):
                # if the localized file does not exist, default to English
                full_path = os.path.join(os.path.dirname(__file__), filename)

        with open(full_path, encoding="UTF-8") as fp:
            rst_source = fp.read()

        if not filename.endswith("index.rst"):
            rst_source = "`" + tr("Home") + " <index.rst>`_\n\n" + rst_source

        self.text.append_rst(rst_source)


def open_help():
    get_workbench().show_view("HelpView")


def load_plugin() -> None:
    get_workbench().add_view(HelpView, tr("Help"), "ne")
    get_workbench().add_command("help_contents", "help", tr("Help contents"), open_help, group=30)



"""
multithreading support.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/_thread.rst.
========================================

.. module:: _thread
   :synopsis: multithreading support

|see_cpython_module| :mod:`python:_thread`.

This module implements multithreading support.

This module is highly experimental and its API is not yet fully settled
and not yet described in this documentation.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver



from pybricks.media.ev3dev import Font, Image, ImageFile
from pybricks.parameters import Color
from typing import Union


class Screen:
    """
    A stub class to represent the screen member of the EV3Brick class.

    Attributes:
        height (int): The height of the screen in pixels.
        width (int): The width of the screen in pixels.
    """

    def __init__(self):
        self.width = 178  # type: int
        self.height = 128  # type: int

    def clear(self):
        """
        Clears the screen. All pixels on the screen will be set to Color.WHITE.
        """
        ...

    def draw_text(self, x: int, y: int, text: str, text_color: Color = Color.BLACK, background_color: Color = None):
        """
        Draws text on the screen.

        The most recent font set using set_font() will be used or Font.DEFAULT if no font has been set yet.

        Args:
            x (int): The x-axis value where the left side of the text will start.
            y (int): The y-axis value where the top of the text will start.
            text (str): The text to draw.
            text_color (Color): The color used for drawing the text.
            background_color (Color): The color used to fill the rectangle behind the text or None for transparent background.
        """
        ...

    def print(self, *args, sep: str = "", end: str = "\n"):
        """
        Prints a line of text on the screen.

        This method works like the builtin print() function, but it writes on the screen instead.

        You can set the font using set_font(). If no font has been set, Font.DEFAULT will be used. The text is always printed used black text with a white background.

        Unlike the builtin print(), the text does not wrap if it is too wide to fit on the screen. It just gets cut off. But if the text would go off of the bottom of the screen, the entire image is scrolled up and the text is printed in the new blank area at the bottom of the screen.

        Args:
            args (object): Zero or more objects to print.
            sep (str): Separator that will be placed between each object that is printed.
            end (str): End of line that will be printed after the last object.
        """
        ...

    def set_font(self, font: Font):
        """
        Sets the font used for writing on the screen.

        The font is used for both draw_text() and print().

        Args:
            font (Font): The font to use.
        """
        ...

    def load_image(self, source: Union[str, Image, ImageFile]):
        """
        Clears this image, then draws the source image centered in the screen.

        Args:
            source (ImageFile, Image, or str): The source Image. If the argument is a string (or ImageFile), then the source image is loaded from file.
        """
        ...

    def draw_image(self, x: int, y: int, source: Union[str, Image, ImageFile], transparent: Color = None):
        """
        Draws the source image on the screen.

        Args:
            x (int): The x-axis value where the left side of the image will start.
            y (int): The y-axis value where the top of the image will start.
            source (ImageFile, Image, str): The source Image. If the argument is a string (or ImageFile), then the source image is loaded from file.
            transparent (Color): The color of image to treat as transparent or None for no transparency.
        """
        ...

    def draw_pixel(self, x: int, y: int, color: Color = Color.BLACK):
        """
        Draws a single pixel on the screen.

        Args:
            x (int): The x coordinate of the pixel.
            y (int): The y coordinate of the pixel.
            color (Color): The color of the pixel.
        """
        ...

    def draw_line(self, x1: int, y1: int, x2: int, y2: int, width: int = 1, color: Color = Color.BLACK):
        """
        Draws a line on the screen.

        Args:
            x1 (int): The x coordinate of the starting point of the line.
            y1 (int): The y coordinate of the starting point of the line.
            x2 (int): The x coordinate of the ending point of the line.
            y2 (int): The y coordinate of the ending point of the line.
            width (int): The width of the line in pixels.
            color (Color): The color of the line.
        """
        ...

    def draw_box(self, x1: int, y1: int, x2: int, y2: int, r: int = 0, fill: bool = False, color: Color = Color.BLACK):
        """
        Draws a box on the screen.

        Args:
            x1 (int): The x coordinate of the left side of the box.
            y1 (int): The y coordinate of the top of the box.
            x2 (int): The x coordinate of the right side of the box.
            y2 (int): The y coordinate of the bottom of the box.
            r (int): The radius of the corners of the box.
            fill (bool): If True, the box will be filled with color, otherwise only the outline of the box will be drawn.
            color (Color): The color of the box.
        """
        ...

    def draw_circle(self, x: int, y: int, r: int, fill: bool = False, color: Color = Color.BLACK):
        """
        Draws a circle on the screen.

        Args:
            x (int): The x coordinate of the center of the circle.
            y (int): The y coordinate of the center of the circle.
            r (int): The radius of the circle.
            fill (bool): If True, the circle will be filled with color, otherwise only the circumference will be drawn.
            color (Color): The color of the circle.
        """
        ...

    def save(self, filename: str):
        """
        Saves the screen as a .png file.

        Args:
            filename (str): The path to the file to be saved.

        Raises:
            TypeError: filename is not a string
            OSError: There was a problem saving the file.
        """
        ...



"""
control the garbage collector.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/gc.rst.
==========================================

.. module:: gc
   :synopsis: control the garbage collector

|see_cpython_module| :mod:`python:gc`.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import overload

def enable() -> None:
    """
   Enable automatic garbage collection.
   """

def disable() -> None:
    """
   Disable automatic garbage collection.  Heap memory can still be allocated,
   and garbage collection can still be initiated manually using :meth:`gc.collect`.
   """

def collect() -> None:
    """
   Run a garbage collection.
   """

def mem_alloc() -> int:
    """
   Return the number of bytes of heap RAM that are allocated.
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This function is MicroPython extension.
   """

def mem_free() -> int:
    """
   Return the number of bytes of available heap RAM, or -1 if this amount
   is not known.
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This function is MicroPython extension.
   """

@overload
def threshold() -> int:
    """
   Set or query the additional GC allocation threshold. Normally, a collection
   is triggered only when a new allocation cannot be satisfied, i.e. on an
   out-of-memory (OOM) condition. If this function is called, in addition to
   OOM, a collection will be triggered each time after *amount* bytes have been
   allocated (in total, since the previous time such an amount of bytes
   have been allocated). *amount* is usually specified as less than the
   full heap size, with the intention to trigger a collection earlier than when the
   heap becomes exhausted, and in the hope that an early collection will prevent
   excessive memory fragmentation. This is a heuristic measure, the effect
   of which will vary from application to application, as well as
   the optimal value of the *amount* parameter.
   
   Calling the function without argument will return the current value of
   the threshold. A value of -1 means a disabled allocation threshold.
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This function is a MicroPython extension. CPython has a similar
      function - ``set_threshold()``, but due to different GC
      implementations, its signature and semantics are different.
   """

@overload
def threshold(amount: int) -> None:
    """
   Set or query the additional GC allocation threshold. Normally, a collection
   is triggered only when a new allocation cannot be satisfied, i.e. on an
   out-of-memory (OOM) condition. If this function is called, in addition to
   OOM, a collection will be triggered each time after *amount* bytes have been
   allocated (in total, since the previous time such an amount of bytes
   have been allocated). *amount* is usually specified as less than the
   full heap size, with the intention to trigger a collection earlier than when the
   heap becomes exhausted, and in the hope that an early collection will prevent
   excessive memory fragmentation. This is a heuristic measure, the effect
   of which will vary from application to application, as well as
   the optimal value of the *amount* parameter.
   
   Calling the function without argument will return the current value of
   the threshold. A value of -1 means a disabled allocation threshold.
   
   .. admonition:: Difference to CPython
      :class: attention
   
      This function is a MicroPython extension. CPython has a similar
      function - ``set_threshold()``, but due to different GC
      implementations, its signature and semantics are different.
   """



import time
import tkinter as tk
from logging import getLogger
from tkinter import messagebox

from turcar import get_runner, get_workbench
from turcar.codeview import SyntaxText
from turcar.common import InlineCommand
from turcar.editor_helpers import get_text_filename
from turcar.languages import tr

logger = getLogger(__name__)


class OccurrencesHighlighter:
    def __init__(self, text):
        self.text: SyntaxText = text
        self._request_scheduled: bool = False
        get_workbench().bind("highlight_occurrences_response", self._handle_response, True)

    def get_positions_for(self, source, line, column):
        raise NotImplementedError()

    def get_positions(self):
        index = self.text.index("insert")

        # ignore if cursor in open string
        if self.text.tag_prevrange("open_string", index) or self.text.tag_prevrange(
            "open_string3", index
        ):
            return set()

        source = self.text.get("1.0", "end")
        index_parts = index.split(".")
        line, column = int(index_parts[0]), int(index_parts[1])

        return self.get_positions_for(source, line, column)

    def trigger(self):
        self._clear()

        if (
            not get_workbench().get_option("view.name_highlighting")
            or not self.text.is_python_text()
        ):
            return

        def consider_request():
            if time.time() - self.text.get_last_operation_time() < 0.3:
                # wait a bit more, there may be more keypresses or cursor location changes coming
                self.text.after(100, consider_request)
            else:
                try:
                    self._request()
                finally:
                    self._request_scheduled = False

        if not self._request_scheduled:
            self._request_scheduled = True
            self.text.after_idle(consider_request)

    def _clear(self) -> None:
        self.text.tag_remove("matched_name", "1.0", "end")

    def _request(self):
        self._clear()
        row, column = map(int, self.text.index("insert").split("."))
        runner = get_runner()
        if not runner or runner.is_running():
            return

        runner.send_command(
            InlineCommand(
                "highlight_occurrences",
                filename=get_text_filename(self.text),
                source=self.text.get("1.0", "end"),
                row=row,
                column=column,
                text_last_operation_time=self.text.get_last_operation_time(),
            )
        )

    def _handle_response(self, msg) -> None:
        error = getattr(msg, "error", None)
        if error:
            messagebox.showerror(tr("Error"), str(error), master=get_workbench())
            return

        if msg.text_last_operation_time != self.text.get_last_operation_time():
            # response arrived too late
            return

        try:
            if len(msg.references) > 1:
                for ref in msg.references:
                    start_index = f"{ref.row}.{ref.column}"
                    end_index = f"{ref.row}.{ref.column + ref.length}"
                    self.text.tag_add("matched_name", start_index, end_index)
        except Exception as e:
            logger.exception("Problem when updating name highlighting", exc_info=e)


def update_highlighting(event):
    if not get_workbench().ready:
        # don't slow down loading process
        return

    if not get_runner() or not get_runner().get_backend_proxy():
        # too early
        return

    assert isinstance(event.widget, tk.Text)
    text = event.widget
    if not hasattr(text, "name_highlighter"):
        text.name_highlighter = OccurrencesHighlighter(text)

    text.name_highlighter.trigger()


def load_plugin() -> None:
    wb = get_workbench()
    wb.set_default("view.name_highlighting", True)
    wb.bind_class("EditorCodeViewText", "<<CursorMove>>", update_highlighting, True)
    wb.bind_class("EditorCodeViewText", "<<TextChange>>", update_highlighting, True)
    wb.bind("<<UpdateAppearance>>", update_highlighting, True)



"""
wait for events on a set of streams.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/select.rst.
====================================================

.. module:: select
   :synopsis: wait for events on a set of streams

|see_cpython_module| :mod:`python:select`.

This module provides functions to efficiently wait for events on multiple
`streams <stream>` (select streams which are ready for operations).
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Iterable, Any, Final, Iterator

from uio import IOBase

POLLIN: Final[int] = ...
"""Data available for reading."""

POLLOUT: Final[int] = ...
"""More data can be written."""

POLLHUP: Final[int] = ...
"""Socket is no longer connected."""

POLLERR: Final[int] = ...
"""Socket got an asynchronous error."""

def poll() -> "Poll":
    """
   Create an instance of the Poll class.
   """

def select(
    rlist: Iterable[Any],
    wlist: Iterable[Any],
    xlist: Iterable[Any],
    timeout: int = -1,
    /,
) -> list[tuple[Any, int, Any, ...]]:
    """
   Wait for activity on a set of objects.
   
   This function is provided by some MicroPython ports for compatibility
   and is not efficient. Usage of :class:`Poll` is recommended instead.
   """

class Poll:
    """
   The name, `Poll`, used for typing is not the same as the runtime name, `poll` (note lowercase `p`).
   The reason for this difference is that the runtime uses `poll` as both a class name and as a method name and
   this is not possible within code written entirely in Python and therefore not possible within typing code.
   """

    def register(self, obj: IOBase, eventmask: int = POLLIN | POLLOUT, /) -> None:
        """
      Register `stream` *obj* for polling. *eventmask* is logical OR of:
      
      * ``select.POLLIN``  - data available for reading
      * ``select.POLLOUT`` - more data can be written
      
      Note that flags like ``select.POLLHUP`` and ``select.POLLERR`` are
      *not* valid as input eventmask (these are unsolicited events which
      will be returned from `poll()` regardless of whether they are asked
      for). This semantics is per POSIX.
      
      *eventmask* defaults to ``select.POLLIN | select.POLLOUT``.
      
      It is OK to call this function multiple times for the same *obj*.
      Successive calls will update *obj*'s eventmask to the value of
      *eventmask* (i.e. will behave as `modify()`).
      """
    def unregister(self, obj: IOBase, /) -> None:
        """
      Unregister *obj* from polling.
      """
    def modify(self, obj: IOBase, eventmask: int, /) -> None:
        """
      Modify the *eventmask* for *obj*. If *obj* is not registered, `OSError`
      is raised with error of ENOENT.
      """
    def poll(self, timeout: int = -1, /) -> list[tuple[Any, int, Any, ...]]:
        """
      Wait for at least one of the registered objects to become ready or have an
      exceptional condition, with optional timeout in milliseconds (if *timeout*
      arg is not specified or -1, there is no timeout).
      
      Returns list of (``obj``, ``event``, ...) tuples. There may be other elements in
      tuple, depending on a platform and version, so don't assume that its size is 2.
      The ``event`` element specifies which events happened with a stream and
      is a combination of ``select.POLL*`` constants described above. Note that
      flags ``select.POLLHUP`` and ``select.POLLERR`` can be returned at any time
      (even if were not asked for), and must be acted on accordingly (the
      corresponding stream unregistered from poll and likely closed), because
      otherwise all further invocations of `poll()` may return immediately with
      these flags set for this stream again.
      
      In case of timeout, an empty list is returned.
      
      .. admonition:: Difference to CPython
         :class: attention
      
         Tuples returned may contain more than 2 elements as described above.
      """
    def ipoll(
        self, timeout: int = -1, flags: int = 0, /
    ) -> Iterator[tuple[Any, int, Any, ...]]:
        """
      Like :meth:`poll.poll`, but instead returns an iterator which yields a
      `callee-owned tuple`. This function provides an efficient, allocation-free
      way to poll on streams.
      
      If *flags* is 1, one-shot behaviour for events is employed: streams for
      which events happened will have their event masks automatically reset
      (equivalent to ``poll.modify(obj, 0)``), so new events for such a stream
      won't be processed until new mask is set with `poll.modify()`. This
      behaviour is useful for asynchronous I/O schedulers.
      
      .. admonition:: Difference to CPython
         :class: attention
      
         This function is a MicroPython extension.
      """



import tkinter as tk

from turcar import get_workbench
from turcar.common import TextRange
from turcar.languages import tr
from turcar.ui_utils import select_sequence

BLOCK_COMMENT_PREFIX = "# "


def _get_focused_writable_text():
    widget = get_workbench().focus_get()
    # In Ubuntu when moving from one menu to another, this may give None when text is actually focused
    if isinstance(widget, tk.Text) and (
        not hasattr(widget, "is_read_only") or not widget.is_read_only()
    ):
        return widget
    else:
        return None


def _writable_text_is_focused():
    return _get_focused_writable_text() is not None


def _selection_is_line_commented(text):
    sel_range = _get_focused_code_range(text)

    for lineno in range(sel_range.lineno, sel_range.end_lineno + 1):
        line = text.get(str(lineno) + ".0", str(lineno) + ".end")
        if not line.startswith(BLOCK_COMMENT_PREFIX):
            return False

    return True


def _select_lines(text, first_line, last_line):
    text.tag_remove("sel", "1.0", tk.END)
    text.tag_add("sel", str(first_line) + ".0", str(last_line) + ".end")


def _toggle_selection_comment(text):
    if _selection_is_line_commented(text):
        _uncomment_selection(text)
    else:
        _comment_selection(text)


def _comment_selection(text):
    """Adds ## in front of all selected lines if any lines are selected,
    or just the current line otherwise"""

    sel_range = _get_focused_code_range(text)

    for lineno in range(sel_range.lineno, sel_range.end_lineno + 1):
        text.insert(str(lineno) + ".0", BLOCK_COMMENT_PREFIX)

    if sel_range.end_lineno > sel_range.lineno:
        _select_lines(text, sel_range.lineno, sel_range.end_lineno)

    text.edit_separator()


def _uncomment_selection(text):
    sel_range = _get_focused_code_range(text)

    for lineno in range(sel_range.lineno, sel_range.end_lineno + 1):
        line = text.get(str(lineno) + ".0", str(lineno) + ".end")
        if line.startswith(BLOCK_COMMENT_PREFIX):
            text.delete(str(lineno) + ".0", str(lineno) + "." + str(len(BLOCK_COMMENT_PREFIX)))


def _get_focused_code_range(text):
    if len(text.tag_ranges("sel")) > 0:
        lineno, col_offset = map(int, text.index(tk.SEL_FIRST).split("."))
        end_lineno, end_col_offset = map(int, text.index(tk.SEL_LAST).split("."))

        if end_lineno > lineno and end_col_offset == 0:
            # SelectAll includes nonexisting extra line
            end_lineno -= 1
            end_col_offset = int(text.index(str(end_lineno) + ".end").split(".")[1])
    else:
        lineno, col_offset = map(int, text.index(tk.INSERT).split("."))
        end_lineno, end_col_offset = lineno, col_offset

    return TextRange(lineno, col_offset, end_lineno, end_col_offset)


def _cmd_toggle_selection_comment():
    text = _get_focused_writable_text()
    if text is not None:
        _toggle_selection_comment(text)


def _cmd_comment_selection():
    text = _get_focused_writable_text()
    if text is not None:
        _comment_selection(text)


def _cmd_uncomment_selection():
    text = _get_focused_writable_text()
    if text is not None:
        _uncomment_selection(text)


def _cmd_indent_selection():
    text = _get_focused_writable_text()
    if text is not None and hasattr(text, "indent_region"):
        text.indent_region()


def _cmd_dedent_selection():
    text = _get_focused_writable_text()
    if text is not None and hasattr(text, "dedent_region"):
        text.dedent_region()


def _cmd_replace_tabs():
    text = _get_focused_writable_text()
    if text is not None:
        orig_lines = text.get("1.0", "end").splitlines(keepends=True)
        new_lines = []
        for line in orig_lines:
            leading_tab_count = 0
            for char in line:
                if char == "\t":
                    leading_tab_count += 1
                else:
                    break
            new_lines.append(leading_tab_count * "    " + line[leading_tab_count:])

        text.delete("1.0", "end")

        text.insert("1.0", "".join(new_lines))


def load_plugin() -> None:
    get_workbench().add_command(
        "indent",
        "edit",
        tr("Indent selected lines"),
        _cmd_indent_selection,
        tester=_writable_text_is_focused,
        accelerator="Tab",
        group=49,
    )

    get_workbench().add_command(
        "dedent",
        "edit",
        tr("Dedent selected lines"),
        _cmd_dedent_selection,
        tester=_writable_text_is_focused,
        accelerator="Shift+Tab",
        group=49,
    )

    get_workbench().add_command(
        "replace_tabs",
        "edit",
        tr("Replace tabs with spaces"),
        _cmd_replace_tabs,
        tester=_writable_text_is_focused,
        group=49,
    )

    get_workbench().add_command(
        "toggle_comment",
        "edit",
        tr("Toggle comment"),
        _cmd_toggle_selection_comment,
        default_sequence=select_sequence("<Control-Key-3>", "<Command-Key-3>"),
        tester=_writable_text_is_focused,
        group=50,
    )

    get_workbench().add_command(
        "comment_selection",
        "edit",
        tr("Comment out"),
        _cmd_comment_selection,
        default_sequence="<Alt-Key-3>",
        tester=_writable_text_is_focused,
        group=50,
    )

    get_workbench().add_command(
        "uncomment_selection",
        "edit",
        tr("Uncomment"),
        _cmd_uncomment_selection,
        default_sequence="<Alt-Key-4>",
        tester=_writable_text_is_focused,
        group=50,
    )



from typing import Tuple

class Control:
    """
    Class to interact with PID controller and settings.

    Attributes:
        scale (float): Scaling factor between the controlled integer variable and the physical output. For example, for a single motor this is the number of encoder pulses per degree of rotation.
    """

    def __init__(self):
        self.scale = 1.0  # type: float

    def done(self) -> bool:
        """
        Checks if an ongoing command or maneuver is done.

        Returns:
            True if the command is done, False if not.
        """
        return False

    def stalled(self) -> bool:
        """
        Checks if the controller is currently stalled.

        A controller is stalled when it cannot reach the target speed or position, even with the maximum actuation signal.

        Returns:
            True if the controller is stalled, False if not.
        """
        return False

    def limits(self, speed: int = None, acceleration: int = None, actuation: int = None) -> Tuple[int, int, int]:
        """
        Configures the maximum speed, acceleration, and actuation.

        If no arguments are given, this will return the current values.

        Args:
            speed (int): Maximum speed. All speed commands will be capped to this value. Rotational (degrees/second) or Linear (millimeters/second).
            acceleration (int): Maximum acceleration. Rotational (degrees/second^2) or Linear (millimeters/second^2).
            actuation (int): Maximum actuation as percentage (0 - 100) of absolute maximum.

        Returns:
            Speed, acceleration, and actuation (if no arguments are provided). None, otherwise.
        """
        if speed is None and acceleration is None and actuation is None:
            return (0, 0, 0)
        else:
            return None

    def pid(self, kp: int = None, ki: int = None, kd: int = None, integral_range: int = None, integral_rate: int = None, feed_forward: int = None) -> Tuple[int, int, int, int, int, int]:
        """
        Gets or sets the PID values for position and speed control.

        If no arguments are given, this will return the current values.

        Args:	
            kp (int): Proportional position (or integral speed) control constant.
            ki (int): Integral position control constant.
            kd (int): Derivative position (or proportional speed) control constant.
            integral_range (int): Region around the target angle (degrees) or distance (millimeters), in which integral control errors are accumulated.
            integral_rate (int): Maximum rate at which the error integral is allowed to grow. Rotational (degrees/second) or Linear (millimeters/second).
            feed_forward (int): This adds a feed forward signal to the PID feedback signal, in the direction of the speed reference. This value is expressed as a percentage (0 - 100) of the absolute maximum duty cycle.

        Returns:
            kp, ki, kd, integral range, integral rate, and feed forward (if no arguments are provided), None otherwise.
        """
        if kp is None and ki is None and kd is None and integral_range is None and integral_rate is None and feed_forward is None:
            return (0, 0, 0, 0, 0, 0)
        else:
            return None

    def target_tolerances(self, speed: int = None, position: int = None) -> Tuple[int, int]:
        """
        Gets or sets the tolerances that say when a maneuver is done.

        If no arguments are given, this will return the current values.

        Args:	
            speed (int): Allowed deviation from zero speed before motion is considered complete. Linear (millimeters/second) or Rotational (degrees/second)
            position (init): Allowed deviation from the target before motion is considered complete. Linear (millimeters) or Rotational (degrees).

        Returns:
            Allowed deviation from zero speed and allowed deviation from the target (if no arguments are provided), None otherwise.
        """
        if speed is None and position is None:
            return (0, 0)
        else:
            return None

    def stall_tolerances(self, speed: int = None, time: int = None) -> Tuple[int, int]:
        """
        Gets or sets stalling tolerances.

        If no arguments are given, this will return the current values.

        Args:	
            speed (int): If the controller cannot reach this speed for some time even with maximum actuation, it is stalled. Rotational (degrees/second) or Linear (millimeters/second).
            time (int): How long the controller has to be below this minimum speed before we say it is stalled in milliseconds.

        Returns:
            Threshold speed and time limit (if no arguments are provided), None otherwise.
        """
        if speed is None and time is None:
            return (0, 0)
        else:
            return None



"""
The `wifi` module provides necessary low-level functionality for managing
wifi connections. Use `socketpool` for communicating over the network."""

from __future__ import annotations

import ipaddress
from typing import Iterable, Iterator, Optional, Union

from circuitpython_typing import ReadableBuffer

radio: Radio
"""Wifi radio used to manage both station and AP modes.
This object is the sole instance of `wifi.Radio`."""

class AuthMode:
    """The authentication protocols used by WiFi."""

    OPEN: object
    """Open network. No authentication required."""

    WEP: object
    """Wired Equivalent Privacy."""

    WPA: object
    """Wireless Protected Access."""

    WPA2: object
    """Wireless Protected Access 2."""

    WPA3: object
    """Wireless Protected Access 3."""

    PSK: object
    """Pre-shared Key. (password)"""

    ENTERPRISE: object
    """Each user has a unique credential."""

class Monitor:
    """For monitoring WiFi packets."""

    def __init__(self, channel: Optional[int] = 1, queue: Optional[int] = 128) -> None:
        """Initialize `wifi.Monitor` singleton.

        :param int channel: The WiFi channel to scan.
        :param int queue: The queue size for buffering the packet.

        """
        ...
    channel: int
    """The WiFi channel to scan."""
    queue: int
    """The queue size for buffering the packet."""
    def deinit(self) -> None:
        """De-initialize `wifi.Monitor` singleton."""
        ...
    def lost(self) -> int:
        """Returns the packet loss count. The counter resets after each poll."""
        ...
    def queued(self) -> int:
        """Returns the packet queued count."""
        ...
    def packet(self) -> dict:
        """Returns the monitor packet."""
        ...

class Network:
    """A wifi network provided by a nearby access point."""

    def __init__(self) -> None:
        """You cannot create an instance of `wifi.Network`. They are returned by `wifi.Radio.start_scanning_networks`."""
        ...
    ssid: str
    """String id of the network"""
    bssid: bytes
    """BSSID of the network (usually the AP's MAC address)"""
    rssi: int
    """Signal strength of the network"""
    channel: int
    """Channel number the network is operating on"""
    country: str
    """String id of the country code"""
    authmode: str
    """String id of the authmode"""

class Packet:
    """The packet parameters."""

    CH: object
    """The packet's channel."""

    LEN: object
    """The packet's length."""

    RAW: object
    """The packet's payload."""

    RSSI: object
    """The packet's rssi."""

class Radio:
    """Native wifi radio.

    This class manages the station and access point functionality of the native
    Wifi radio.

    """

    def __init__(self) -> None:
        """You cannot create an instance of `wifi.Radio`.
        Use `wifi.radio` to access the sole instance available."""
        ...
    enabled: bool
    """``True`` when the wifi radio is enabled.
    If you set the value to ``False``, any open sockets will be closed.
    """
    hostname: Union[str | ReadableBuffer]
    """Hostname for wifi interface. When the hostname is altered after interface started/connected
       the changes would only be reflected once the interface restarts/reconnects."""
    mac_address: ReadableBuffer
    """MAC address for the station. When the address is altered after interface is connected
       the changes would only be reflected once the interface reconnects.

    **Limitations:** Not settable on RP2040 CYW43 boards, such as Pi Pico W.
    """
    tx_power: float
    """Wifi transmission power, in dBm."""
    mac_address_ap: ReadableBuffer
    """MAC address for the AP. When the address is altered after interface is started
       the changes would only be reflected once the interface restarts.

    **Limitations:** Not settable on RP2040 CYW43 boards, such as Pi Pico W.
    """
    def start_scanning_networks(
        self, *, start_channel: int = 1, stop_channel: int = 11
    ) -> Iterable[Network]:
        """Scans for available wifi networks over the given channel range. Make sure the channels are allowed in your country.

        .. note::

            In the raspberrypi port (RP2040 CYW43), ``start_channel`` and ``stop_channel`` are ignored.
        """
        ...
    def stop_scanning_networks(self) -> None:
        """Stop scanning for Wifi networks and free any resources used to do it."""
        ...
    def start_station(self) -> None:
        """Starts a Station."""
        ...
    def stop_station(self) -> None:
        """Stops the Station."""
        ...
    def start_ap(
        self,
        ssid: Union[str | ReadableBuffer],
        password: Union[str | ReadableBuffer] = b"",
        *,
        channel: int = 1,
        authmode: Optional[AuthMode] = None,
        max_connections: Optional[int] = 4,
    ) -> None:
        """Starts an Access Point with the specified ssid and password.

        If ``channel`` is given, the access point will use that channel unless
        a station is already operating on a different channel.

        If ``authmode`` is not None, the access point will use that Authentication
        mode. If a non-empty password is given, ``authmode`` must not be ``OPEN``.
        If ``authmode`` is not given or is None,
        ``OPEN`` will be used when the password is the empty string,
        otherwise ``authmode`` will be ``WPA_WPA2_PSK``.

        If ``max_connections`` is given, the access point will allow up to
        that number of stations to connect."""
        ...
    def stop_ap(self) -> None:
        """Stops the Access Point."""
        ...
    def connect(
        self,
        ssid: Union[str | ReadableBuffer],
        password: Union[str | ReadableBuffer] = b"",
        *,
        channel: int = 0,
        bssid: Optional[Union[str | ReadableBuffer]] = None,
        timeout: Optional[float] = None,
    ) -> None:
        """Connects to the given ssid and waits for an ip address. Reconnections are handled
        automatically once one connection succeeds.

        By default, this will scan all channels and connect to the access point (AP) with the
        given ``ssid`` and greatest signal strength (rssi).

        If ``channel`` is non-zero, the scan will begin with the given channel and connect to
        the first AP with the given ``ssid``. This can speed up the connection time
        significantly because a full scan doesn't occur.

        If ``bssid`` is given and not None, the scan will start at the first channel or the one given and
        connect to the AP with the given ``bssid`` and ``ssid``."""
        ...
    ipv4_gateway: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the station gateway when connected to an access point. None otherwise."""
    ipv4_gateway_ap: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the access point gateway, when enabled. None otherwise."""
    ipv4_subnet: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the station subnet when connected to an access point. None otherwise."""
    ipv4_subnet_ap: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the access point subnet, when enabled. None otherwise."""
    def set_ipv4_address(
        self,
        *,
        ipv4: ipaddress.IPv4Address,
        netmask: ipaddress.IPv4Address,
        gateway: ipaddress.IPv4Address,
        ipv4_dns: Optional[ipaddress.IPv4Address],
    ) -> None:
        """Sets the IP v4 address of the station. Must include the netmask and gateway. DNS address is optional.
        Setting the address manually will stop the DHCP client."""
        ...
    ipv4_address: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the station when connected to an access point. None otherwise."""
    ipv4_address_ap: Optional[ipaddress.IPv4Address]
    """IP v4 Address of the access point, when enabled. None otherwise."""
    ipv4_dns: ipaddress.IPv4Address
    """IP v4 Address of the DNS server to be used."""
    ap_info: Optional[Network]
    """Network object containing BSSID, SSID, authmode, channel, country and RSSI when connected to an access point. None otherwise."""
    def start_dhcp(self) -> None:
        """Starts the DHCP client."""
        ...
    def stop_dhcp(self) -> None:
        """Stops the DHCP client. Needed to assign a static IP address."""
        ...
    def ping(
        self, ip: ipaddress.IPv4Address, *, timeout: Optional[float] = 0.5
    ) -> Optional[float]:
        """Ping an IP to test connectivity. Returns echo time in seconds.
        Returns None when it times out."""
        ...

class ScannedNetworks:
    """Iterates over all `wifi.Network` objects found while scanning. This object is always created
    by a `wifi.Radio`: it has no user-visible constructor."""

    def __init__(self) -> None:
        """Cannot be instantiated directly. Use `wifi.Radio.start_scanning_networks`."""
        ...
    def __iter__(self) -> Iterator[Network]:
        """Returns itself since it is the iterator."""
        ...
    def __next__(self) -> Network:
        """Returns the next `wifi.Network`.
        Raises `StopIteration` if scanning is finished and no other results are available.
        """
        ...



"""Watchdog Timer

The `watchdog` module provides support for a Watchdog Timer. This timer will reset the device
if it hasn't been fed after a specified amount of time. This is useful to ensure the board
has not crashed or locked up. Note that on some platforms the watchdog timer cannot be disabled
once it has been enabled.

The `WatchDogTimer` is used to restart the system when the application crashes and ends
up into a non recoverable state. Once started it cannot be stopped or
reconfigured in any way. After enabling, the application must "feed" the
watchdog periodically to prevent it from expiring and resetting the system.

Example usage::

    from microcontroller import watchdog as w
    from watchdog import WatchDogMode
    w.timeout=2.5 # Set a timeout of 2.5 seconds
    w.mode = WatchDogMode.RAISE
    w.feed()"""

from __future__ import annotations

class WatchDogTimeout(Exception):
    """Exception raised when the watchdog timer is set to
    ``WatchDogMode.RAISE`` and expires.

    Example::

        import microcontroller
        import watchdog
        import time

        wdt = microcontroller.watchdog
        wdt.timeout = 5

        while True:
            wdt.mode = watchdog.WatchDogMode.RAISE
            print("Starting loop -- should exit after five seconds")
            try:
                while True:
                    time.sleep(10)  # Also works with pass
            except watchdog.WatchDogTimeout as e:
                print("Watchdog expired")
            except Exception as e:
                print("Other exception")

        print("Exited loop")
    """

class WatchDogMode:
    """run state of the watchdog timer"""

    def __init__(self) -> None:
        """Enum-like class to define the run mode of the watchdog timer."""
    RAISE: WatchDogMode
    """Raise an exception when the WatchDogTimer expires.

    **Limitations:** ``RAISE`` mode is not supported on SAMD or RP2040.

    :type WatchDogMode:"""

    RESET: WatchDogMode
    """Reset the system if the WatchDogTimer expires.

    :type WatchDogMode:"""

class WatchDogTimer:
    """Timer that is used to detect code lock ups and automatically reset the microcontroller
    when one is detected.

    A lock up is detected when the watchdog hasn't been fed after a given duration. So, make
    sure to call `feed` within the timeout.
    """

    def __init__(self) -> None:
        """Not currently dynamically supported. Access the sole instance through `microcontroller.watchdog`."""
        ...
    def feed(self) -> None:
        """Feed the watchdog timer. This must be called regularly, otherwise
        the timer will expire."""
        ...
    def deinit(self) -> None:
        """Stop the watchdog timer. This may raise an error if the watchdog
        timer cannot be disabled on this platform."""
        ...
    timeout: float
    """The maximum number of seconds that can elapse between calls
    to feed()"""
    mode: WatchDogMode
    """The current operating mode of the WatchDogTimer `watchdog.WatchDogMode`.

    Setting a WatchDogMode activates the WatchDog::

      import microcontroller
      import watchdog

      w = microcontroller.watchdog
      w.timeout = 5
      w.mode = watchdog.WatchDogMode.RAISE


    Once set, the WatchDogTimer will perform the specified action if the timer expires."""



"""
Module: 'mindstorms.operator' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2

from _api.operator import equal_to, greater_than, greater_than_or_equal_to, less_than, less_than_or_equal_to, not_equal_to



"""
The lower-level hub API.

Described at https://lego.github.io/MINDSTORMS-Robot-Inventor-hub-API/
"""
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterable,
    List,
    Optional,
    Protocol,
    Tuple,
    Union,
    overload,
)

__version__: str
"""
The firmware version of the form 'v1.0.06.0034-b0c335b', consisting of the components:
v major . minor . bugfix . build - hash.
"""

config: Dict[str, Any] = {}

# Orientation constants
TOP: Final[int] = 0
FRONT: Final[int] = 1
RIGHT: Final[int] = 2
BOTTOM: Final[int] = 3
BACK: Final[int] = 4
LEFT: Final[int] = 5

class _Button(Protocol):
    """
    Provides access to button state and callback.

    Each of the buttons listed in `hub.button` are instances of this class. You cannot instantiate additional button objects.
    """

    def is_pressed(self) -> bool:
        """
        Gets the state of the button.
        """
    def was_pressed(self) -> bool:
        """
        Checks if this button was pressed since this method was last called.
        """
    def presses(self) -> int:
        """
        Gets the number of times this button was pressed since this method was last called.
        """
    def on_change(self, function: Optional[Callable[[int], None]]) -> Any:
        """TODO: Same as `callback`?"""
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """
        Sets the callback function that is called when the button is pressed and when it is released.
        Choose None to disable the callback.

        The function must accept one argument, whose value indicates why the callback was called.
        If the value is 0, the button is now pressed.
        Otherwise, the button is now released. The value represents how many milliseconds it was pressed before it was released.
        """

class _Device(Protocol):
    """
    Read values from a Powered Up device and configure its modes.

    This class is the blueprint for the device attributes of the ports in the ``hub.port`` module, which in turn are instances of the Port class. You cannot import or instantiate this class manually.
    """

    # TODO:

class _Motor(Protocol):
    """
    TODO:
    """

class _MotorPair(Protocol):
    """
    TODO:
    """

class _Pin(Protocol):
    def direction(self, direction: Optional[int]) -> int:
        """"""
    def value(self, value: Optional[int]) -> int:
        """"""

class _Port(Protocol):
    """
    Provides access to port configuration and devices on a port. Some methods and attributes can only be used if the port is in the right mode, as shown below.

    This class is the blueprint for the port instances in the ``hub.port`` module. Those instances are automatically instantiated on boot, and further populated when devices are plugged in. You cannot import or instantiate this class manually.
    """

    device: Optional[_Device]
    motor: Optional[_Motor]
    p5: _Pin
    p6: _Pin

    def pwm(self, value: int) -> None:
        """"""
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""
    def mode(self, mode: int, baud_rate: int = 2400) -> None:
        """"""
    def info(self) -> Dict[str, Any]:
        """"""
    def baud(self, baud: int) -> None:
        """"""
    def read(self, read: Union[int, Any]) -> int:
        """"""
    def write(self, data: bytes) -> int:
        """"""

class _battery(Protocol):
    """
    Battery-related constants and functions.
    """

    BATTERY_BAD_BATTERY: Final[int] = -4
    BATTERY_HUB_TEMPERATURE_CRITICAL_OUT_OF_RANGE: Final[int] = -2
    BATTERY_MISSING: Final[int] = -6
    BATTERY_NO_ERROR: Final[int] = 0
    BATTERY_TEMPERATURE_OUT_OF_RANGE: Final[int] = -2
    BATTERY_TEMPERATURE_SENSOR_FAIL: Final[int] = -3
    BATTERY_VOLTAGE_TOO_LOW: Final[int] = -5
    CHARGER_STATE_CHARGING_COMPLETED: Final[int] = 2
    CHARGER_STATE_CHARGING_ONGOING: Final[int] = 1
    CHARGER_STATE_DISCHARGING: Final[int] = 0
    CHARGER_STATE_FAIL: Final[int] = -1
    USB_CH_PORT_CDP: Final[int] = 2
    USB_CH_PORT_DCP: Final[int] = 3
    USB_CH_PORT_NONE: Final[int] = 0
    USB_CH_PORT_SDP: Final[int] = 1

    def voltage(self) -> int:
        """"""
    def current(self) -> int:
        """"""
    def capacity_left(self) -> int:
        """"""
    def temperature(self) -> float:
        """"""
    def info(self) -> Dict[str, Any]:
        """"""
    def charger_detect(self) -> Union[bool, int]:
        """"""

class _bluetooth(Protocol):
    @overload
    def discoverable(self) -> int:
        """"""
    @overload
    def discoverable(self, time: int) -> None:
        """"""
    @overload
    def rfcomm_connect(self) -> str:
        """"""
    @overload
    def rfcomm_connect(self, address: str) -> None:
        """"""
    def rfcomm_disconnect(self) -> None:
        """"""
    def info(self) -> Dict[str, Any]:
        """"""
    def forget(self, address: str) -> bool:
        """"""
    @overload
    def lwp_advertise(self) -> int:
        """"""
    @overload
    def lwp_advertise(self, bypass: bool) -> None:
        """"""
    @overload
    def lwp_bypass(self) -> bool:
        """"""
    @overload
    def lwp_bypass(self, bypass: bool) -> None:
        """"""
    def lwp_monitor(self, handler: Optional[Callable[[int], None]]) -> None:
        """"""

class _button(Protocol):
    """
    Button instances.
    """

    center: _Button
    connect: _Button
    left: _Button
    right: _Button

class _display(Protocol):
    def clear(self) -> None:
        """"""
    def rotation(self, rotation: int) -> None:
        """"""
    @overload
    def align(self) -> int:
        """"""
    @overload
    def align(self, face: int) -> int:
        """"""
    @overload
    def invert(self) -> bool:
        """"""
    @overload
    def invert(self, invert: bool) -> bool:
        """"""
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""
    @overload
    def pixel(self, x: int, y: int) -> int:
        """"""
    @overload
    def pixel(self, x: int, y: int, brightness: int) -> None:
        """"""
    @overload
    def show(self, image: Image) -> None:
        """"""
    @overload
    def show(
        self,
        image: Iterable[Image],
        delay: int = 400,
        level: int = 9,
        clear: bool = False,
        wait: bool = True,
        loop: bool = False,
        fade: int = 0,
    ) -> None:
        """"""

class _motion(Protocol):
    DOUBLETAPPED: Final[int] = 1
    FREEFALL: Final[int] = 3
    SHAKE: Final[int] = 2
    TAPPED: Final[int] = 0

    def accelerometer(self, filtered: bool = False) -> Tuple[int, int, int]:
        """"""
    def gyroscope(self, filtered: bool = False) -> Tuple[int, int, int]:
        """"""
    @overload
    def align_to_model(self, top: int, front: int) -> None:
        """"""
    @overload
    def align_to_model(self, nsamples: int, callback: Callable[[int], None]) -> None:
        """"""
    @overload
    def align_to_model(
        self, top: int, front: int, nsamples: int, callback: Callable[[int], None]
    ) -> None:
        """"""
    @overload
    def yaw_pitch_roll(self) -> Tuple[int, int, int]:
        """"""
    @overload
    def yaw_pitch_roll(self, yaw_preset: int) -> None:
        """"""
    @overload
    def yaw_pitch_roll(self, yaw_correction: float) -> None:
        """"""
    @overload
    def orientation(self) -> int:
        """"""
    @overload
    def orientation(self, callback: Optional[Callable[[int], None]]) -> int:
        """"""
    @overload
    def gesture(self) -> int:
        """"""
    @overload
    def gesture(self, callback: Optional[Callable[[int], None]]) -> int:
        """"""

class _port(Protocol):
    A: _Port
    B: _Port
    C: _Port
    D: _Port
    E: _Port
    F: _Port

    DETACHED: Final[int] = 0
    ATTACHED: Final[int] = 1

    MODE_DEFAULT: Final[int] = 0
    MODE_FULL_DUPLEX: Final[int] = 1
    MODE_HALF_DUPLEX: Final[int] = 2
    MODE_GPIO: Final[int] = 3

class _sound(Protocol):
    SOUND_SIN: Final[int] = 0
    SOUND_SQUARE: Final[int] = 1
    SOUND_TRIANGLE: Final[int] = 2
    SOUND_SAWTOOTH: Final[int] = 3

    @overload
    def volume(self, volume: int) -> None:
        """"""
    @overload
    def volume(self) -> int:
        """"""
    def beep(self, freq: int = 1000, time: int = 1000, waveform: int = 0) -> None:
        """"""
    def play(self, filename: str, rate: int = 16000) -> None:
        """"""
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""

class _supervision(Protocol):
    def info(self) -> Dict[str, Any]:
        """"""
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""

class Image:
    ALL_ARROWS: Tuple[Image, Image, Image, Image, Image, Image, Image, Image]
    ALL_CLOCKS: Tuple[
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
        Image,
    ]
    ANGRY: Image  ## <class 'Image'> = Image('90009:09090:00000:99999:90909:')
    ARROW_E: Image  ## <class 'Image'> = Image('00900:00090:99999:00090:00900:')
    ARROW_N: Image  ## <class 'Image'> = Image('00900:09990:90909:00900:00900:')
    ARROW_NE: Image  ## <class 'Image'> = Image('00999:00099:00909:09000:90000:')
    ARROW_NW: Image  ## <class 'Image'> = Image('99900:99000:90900:00090:00009:')
    ARROW_S: Image  ## <class 'Image'> = Image('00900:00900:90909:09990:00900:')
    ARROW_SE: Image  ## <class 'Image'> = Image('90000:09000:00909:00099:00999:')
    ARROW_SW: Image  ## <class 'Image'> = Image('00009:00090:90900:99000:99900:')
    ARROW_W: Image  ## <class 'Image'> = Image('00900:09000:99999:09000:00900:')
    ASLEEP: Image  ## <class 'Image'> = Image('00000:99099:00000:09990:00000:')
    BUTTERFLY: Image  ## <class 'Image'> = Image('99099:99999:00900:99999:99099:')
    CHESSBOARD: Image  ## <class 'Image'> = Image('09090:90909:09090:90909:09090:')
    CLOCK1: Image  ## <class 'Image'> = Image('00090:00090:00900:00000:00000:')
    CLOCK10: Image  ## <class 'Image'> = Image('00000:99000:00900:00000:00000:')
    CLOCK11: Image  ## <class 'Image'> = Image('09000:09000:00900:00000:00000:')
    CLOCK12: Image  ## <class 'Image'> = Image('00900:00900:00900:00000:00000:')
    CLOCK2: Image  ## <class 'Image'> = Image('00000:00099:00900:00000:00000:')
    CLOCK3: Image  ## <class 'Image'> = Image('00000:00000:00999:00000:00000:')
    CLOCK4: Image  ## <class 'Image'> = Image('00000:00000:00900:00099:00000:')
    CLOCK5: Image  ## <class 'Image'> = Image('00000:00000:00900:00090:00090:')
    CLOCK6: Image  ## <class 'Image'> = Image('00000:00000:00900:00900:00900:')
    CLOCK7: Image  ## <class 'Image'> = Image('00000:00000:00900:09000:09000:')
    CLOCK8: Image  ## <class 'Image'> = Image('00000:00000:00900:99000:00000:')
    CLOCK9: Image  ## <class 'Image'> = Image('00000:00000:99900:00000:00000:')
    CONFUSED: Image  ## <class 'Image'> = Image('00000:09090:00000:09090:90909:')
    COW: Image  ## <class 'Image'> = Image('90009:90009:99999:09990:00900:')
    DIAMOND: Image  ## <class 'Image'> = Image('00900:09090:90009:09090:00900:')
    DIAMOND_SMALL: Image  ## <class 'Image'> = Image('00000:00900:09090:00900:00000:')
    DUCK: Image  ## <class 'Image'> = Image('09900:99900:09999:09990:00000:')
    FABULOUS: Image  ## <class 'Image'> = Image('99999:99099:00000:09090:09990:')
    GHOST: Image  ## <class 'Image'> = Image('99999:90909:99999:99999:90909:')
    GIRAFFE: Image  ## <class 'Image'> = Image('99000:09000:09000:09990:09090:')
    GO_DOWN: Image  ## <class 'Image'> = Image('00000:99999:09990:00900:00000:')
    GO_LEFT: Image  ## <class 'Image'> = Image('00090:00990:09990:00990:00090:')
    GO_RIGHT: Image  ## <class 'Image'> = Image('09000:09900:09990:09900:09000:')
    GO_UP: Image  ## <class 'Image'> = Image('00000:00900:09990:99999:00000:')
    HAPPY: Image  ## <class 'Image'> = Image('00000:09090:00000:90009:09990:')
    HEART: Image  ## <class 'Image'> = Image('09090:99999:99999:09990:00900:')
    HEART_SMALL: Image  ## <class 'Image'> = Image('00000:09090:09990:00900:00000:')
    HOUSE: Image  ## <class 'Image'> = Image('00900:09990:99999:09990:09090:')
    MEH: Image  ## <class 'Image'> = Image('09090:00000:00090:00900:09000:')
    MUSIC_CROTCHET: Image  ## <class 'Image'> = Image('00900:00900:00900:99900:99900:')
    MUSIC_QUAVER: Image  ## <class 'Image'> = Image('00900:00990:00909:99900:99900:')
    MUSIC_QUAVERS: Image  ## <class 'Image'> = Image('09999:09009:09009:99099:99099:')
    NO: Image  ## <class 'Image'> = Image('90009:09090:00900:09090:90009:')
    PACMAN: Image  ## <class 'Image'> = Image('09999:99090:99900:99990:09999:')
    PITCHFORK: Image  ## <class 'Image'> = Image('90909:90909:99999:00900:00900:')
    RABBIT: Image  ## <class 'Image'> = Image('90900:90900:99990:99090:99990:')
    ROLLERSKATE: Image  ## <class 'Image'> = Image('00099:00099:99999:99999:09090:')
    SAD: Image  ## <class 'Image'> = Image('00000:09090:00000:09990:90009:')
    SILLY: Image  ## <class 'Image'> = Image('90009:00000:99999:00909:00999:')
    SKULL: Image  ## <class 'Image'> = Image('09990:90909:99999:09990:09990:')
    SMILE: Image  ## <class 'Image'> = Image('00000:00000:00000:90009:09990:')
    SNAKE: Image  ## <class 'Image'> = Image('99000:99099:09090:09990:00000:')
    SQUARE: Image  ## <class 'Image'> = Image('99999:90009:90009:90009:99999:')
    SQUARE_SMALL: Image  ## <class 'Image'> = Image('00000:09990:09090:09990:00000:')
    STICKFIGURE: Image  ## <class 'Image'> = Image('00900:99999:00900:09090:90009:')
    SURPRISED: Image  ## <class 'Image'> = Image('09090:00000:00900:09090:00900:')
    SWORD: Image  ## <class 'Image'> = Image('00900:00900:00900:09990:00900:')
    TARGET: Image  ## <class 'Image'> = Image('00900:09990:99099:09990:00900:')
    TORTOISE: Image  ## <class 'Image'> = Image('00000:09990:99999:09090:00000:')
    TRIANGLE: Image  ## <class 'Image'> = Image('00000:00900:09090:99999:00000:')
    TRIANGLE_LEFT: Image  ## <class 'Image'> = Image('90000:99000:90900:90090:99999:')
    TSHIRT: Image  ## <class 'Image'> = Image('99099:99999:09990:09990:09990:')
    UMBRELLA: Image  ## <class 'Image'> = Image('09990:99999:00900:90900:09900:')
    XMAS: Image  ## <class 'Image'> = Image('00900:09990:00900:09990:99999:')
    YES: Image  ## <class 'Image'> = Image('00000:00009:00090:90900:09000:')

    @overload
    def __init__(self, string: str) -> None:
        """"""
    @overload
    def __init__(self, width: int, height: int) -> None:
        """"""
    @overload
    def __init__(self, width: int, height: int, buffer: bytes) -> None:
        """"""
    def width(self) -> int:
        """"""
    def height(self) -> int:
        """"""
    def shift_left(self, n: int) -> Image:
        """"""
    def shift_right(self, n: int) -> Image:
        """"""
    def shift_up(self, n: int) -> Image:
        """"""
    def shift_down(self, n: int) -> Image:
        """"""
    def get_pixel(self, x: int, y: int) -> int:
        """"""
    def set_pixel(self, x: int, y: int, brightness: int) -> None:
        """"""

class BT_VCP:
    def __init__(self, id: int = 0) -> None:
        """"""
    def setinterrupt(self, chr: int) -> None:
        """"""
    def isconnected(self) -> bool:
        """"""
    def any(self) -> bool:
        """"""
    def close(self) -> None:
        """"""
    @overload
    def read(self) -> Optional[bytes]:
        """"""
    @overload
    def read(self, nbytes: int) -> Optional[bytes]:
        """"""
    @overload
    def readinto(self, buf) -> Optional[int]:
        """"""
    @overload
    def readinto(self, buf: bytearray, maxlen: int) -> Optional[int]:
        """"""
    def readline(self) -> Optional[bytes]:
        """"""
    def readlines(self, *args, **kwargs) -> List[bytes]:
        """"""
    def write(self, buf: bytes) -> int:
        """"""
    def recv(self, data: Union[int, bytearray], timeout: int = 5000) -> int:
        """"""
    def send(self, data: Union[int, bytes], timeout: int = 5000) -> int:
        """"""
    @overload
    def callback(self) -> Optional[Callable[[int], None]]:
        """"""
    @overload
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""

class USB_VCP:
    RTS: Final[int] = 1
    CTS: Final[int] = 2

    def __init__(self, id: int = 0) -> None:
        """"""
    def init(self, flow: int = -1) -> None:
        """"""
    def setinterrupt(self, chr: int) -> None:
        """"""
    def isconnected(self) -> bool:
        """"""
    def any(self) -> bool:
        """"""
    def close(self) -> None:
        """"""
    @overload
    def read(self) -> Optional[bytes]:
        """"""
    @overload
    def read(self, nbytes: int) -> Optional[bytes]:
        """"""
    @overload
    def readinto(self, buf) -> Optional[int]:
        """"""
    @overload
    def readinto(self, buf: bytearray, maxlen: int) -> Optional[int]:
        """"""
    def readline(self) -> Optional[bytes]:
        """"""
    def readlines(self, *args, **kwargs) -> List[bytes]:
        """"""
    def write(self, buf: bytes) -> int:
        """"""
    def recv(self, data: Union[int, bytearray], timeout: int = 5000) -> int:
        """"""
    def send(self, data: Union[int, bytes], timeout: int = 5000) -> int:
        """"""
    @overload
    def callback(self) -> Optional[Callable[[int], None]]:
        """"""
    @overload
    def callback(self, function: Optional[Callable[[int], None]]) -> None:
        """"""

battery: _battery

bluetooth: _bluetooth

button: _button

display: _display

motion: _motion

port: _port

sound: _sound

supervision: _supervision

def info() -> Dict[str, Any]:
    """Returns information about the hub"""

def status() -> Dict[str, Any]:
    """Returns status of the components."""

@overload
def power_off(fast: bool = True, restart: bool = False) -> None:
    """
    Turns the hub off.

    Use fast=True for fast shut down, without the usual light animation and sound.
    Use restart=True to reboot after shutting down.
    """

@overload
def power_off(timeout: int = 0) -> None:
    """Sets the inactivity timeout before the hub shuts down automatically."""

def repl_restart(restart: bool = True) -> None:
    """
    Resets the REPL and clears all variables if restart=True.

    Does nothing if restart=False.
    """

def temperature() -> float:
    """Returns hub's temperature in Celsius degrees."""

@overload
def led(color: int, /) -> None:
    """Sets the color of the LED in the center button of the hub.

    0 = off
    1 = pink
    2 = violet
    3 = blue
    4 = turquoise
    5 = light green
    6 = green
    7 = yellow
    8 = orange
    9 = red
    10 = white
    """

@overload
def led(red: int, green: int, blue: int, /) -> None:
    """Sets the color of the LED in the center button of the hub as RGB components (0-255)."""

@overload
def led(color: Tuple[int, int, int], /) -> None:
    """Sets the color of the LED in the center button of the hub as tuple of RGB components (0-255)."""

def file_transfer(
    filename: str,
    filesize: int,
    packetsize: int = 1000,
    timeout: int = 2000,
    mode: Optional[str] = None,
) -> None:
    """Prepares a file transfer to the hub."""

def reset() -> None:
    """Resets the hub."""



from pybricks.parameters import Color, Port
from typing import Tuple


class ColorSensor:
    """
    LEGO® MINDSTORMS® EV3 Color Sensor.

    Args:
        port (Port): Port to which the sensor is connected.
    """

    def __init__(self, port: Port):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def color(self) -> Color:
        """
        Measure the color of a surface.

        Returns:
            Color.BLACK, Color.BLUE, Color.GREEN, Color.YELLOW, Color.RED, Color.WHITE Color.BROWN, or None if no color is detected.
        """
        return Color.BLACK

    def ambient(self) -> int:
        """
        Measures the ambient light intensity.

        Returns:
            Ambient light intensity, ranging from 0 (dark) to 100 (bright).
        """
        return 0

    def reflection(self) -> int:
        """
        Measures the reflection of a surface using a red light.

        Returns:
            Reflection, ranging from 0 (no reflection) to 100 (high reflection).
        """
        return 0

    def rgb(self) -> Tuple[int, int, int]:
        """
        Measures the reflection of a surface using a red, green, and then a blue light.

        Returns:
            Tuple of reflections for red, green, and blue light, each ranging from 0 (no reflection) to 100 (high reflection).
        """
        return (0, 0, 0)



import os.path
import subprocess
import sys
import textwrap
import tkinter as tk
import traceback
from logging import getLogger
from tkinter import messagebox, ttk
from typing import Any, Dict

import turcar
from turcar import get_runner, get_shell, get_workbench, running, ui_utils
from turcar.common import (
    InlineCommand,
    InlineResponse,
    ToplevelCommand,
    get_base_executable,
    is_private_python,
    is_virtual_executable,
    normpath_with_actual_case,
)
from turcar.languages import tr
from turcar.misc_utils import running_on_mac_os, running_on_windows
from turcar.plugins.backend_config_page import BackendDetailsConfigPage
from turcar.running import WINDOWS_EXE, SubprocessProxy, get_front_interpreter_for_subprocess
from turcar.terminal import run_in_terminal
from turcar.ui_utils import askdirectory, askopenfilename, create_string_var

logger = getLogger(__name__)


class LocalCPythonProxy(SubprocessProxy):
    def __init__(self, clean: bool) -> None:
        logger.info("Creating LocalCPythonProxy")
        executable = get_workbench().get_option("LocalCPython.executable")
        self._expecting_response_for_gui_update = False
        super().__init__(clean, executable)
        try:
            self._send_msg(ToplevelCommand("get_environment_info"))
        except Exception:
            get_shell().report_exception()

    def _get_initial_cwd(self):
        return get_workbench().get_local_cwd()

    def _get_launch_cwd(self):
        # use a directory which doesn't contain misleading modules
        empty_dir = os.path.join(turcar.turcar_USER_DIR, "leave_this_empty")
        os.makedirs(empty_dir, exist_ok=True)
        return empty_dir

    def _get_launcher_with_args(self):
        launcher_file = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "cpython_backend", "cp_launcher.py"
        )
        return [
            launcher_file,
            self.get_cwd(),
            repr(
                {
                    "run.warn_module_shadowing": get_workbench().get_option(
                        "run.warn_module_shadowing"
                    )
                }
            ),
        ]

    def can_be_isolated(self) -> bool:
        # Can't run in isolated mode as it would hide user site-packages
        return False

    def _store_state_info(self, msg):
        super()._store_state_info(msg)

        if "gui_is_active" in msg:
            self._update_gui_updating(msg)

    def _prepare_clean_launch(self):
        self._close_backend()
        self._start_background_process()

    def _close_backend(self):
        self._cancel_gui_update_loop()
        super()._close_backend()

    def get_target_executable(self):
        return self._mgmt_executable

    def _update_gui_updating(self, msg):
        """Enables running Tkinter or Qt programs which doesn't call mainloop.

        When mainloop is omitted, then program can be interacted with
        from the shell after it runs to the end.

        Each ToplevelResponse is supposed to tell, whether gui is active
        and needs updating.
        """
        if "gui_is_active" not in msg:
            return

        if msg["gui_is_active"] and self._gui_update_loop_id is None:
            # Start updating
            logger.info("Starting GUI update loop")
            self._loop_gui_update(True)
        elif not msg["gui_is_active"] and self._gui_update_loop_id is not None:
            logger.info("Cancelling GUI update loop")
            self._cancel_gui_update_loop()

    def _loop_gui_update(self, force=False):
        if force or get_runner().is_waiting_toplevel_command():
            # Don't send command if response for the last one hasn't arrived yet
            if not self._expecting_response_for_gui_update:
                try:
                    self.send_command(InlineCommand("process_gui_events"))
                    self._expecting_response_for_gui_update = True
                except OSError:
                    # the backend process may have been closed already
                    # https://github.com/turcar/turcar/issues/966
                    logger.exception("Could not send process_gui_events")

        self._gui_update_loop_id = get_workbench().after(50, self._loop_gui_update)

    def _cancel_gui_update_loop(self):
        if self._gui_update_loop_id is not None:
            try:
                get_workbench().after_cancel(self._gui_update_loop_id)
            finally:
                self._gui_update_loop_id = None

    def interrupt(self):
        import signal

        if self._proc is not None and self._proc.poll() is None:
            if running_on_windows():
                try:
                    os.kill(self._proc.pid, signal.CTRL_BREAK_EVENT)  # pylint: disable=no-member
                except Exception:
                    logger.exception("Could not interrupt backend process")
            else:
                self._proc.send_signal(signal.SIGINT)

    def run_script_in_terminal(self, script_path, args, interactive, keep_open):
        cmd = [self._mgmt_executable]
        if interactive:
            cmd.append("-i")
        cmd.append(os.path.basename(script_path))
        cmd.extend(args)

        run_in_terminal(cmd, os.path.dirname(script_path), keep_open=keep_open)

    def get_pip_gui_class(self):
        from turcar.plugins.cpython_frontend.cp_pip_gui import LocalCPythonPipDialog

        return LocalCPythonPipDialog

    def can_run_remote_files(self):
        return False

    def can_run_local_files(self):
        return True

    def fetch_next_message(self):
        while True:
            msg = super().fetch_next_message()
            if isinstance(msg, InlineResponse) and msg.command_name == "process_gui_events":
                # Only wanted to know that the command was processed
                # Don't pass upstream
                self._expecting_response_for_gui_update = False
            else:
                break

        return msg

    def has_local_interpreter(self):
        return True

    def can_debug(self) -> bool:
        return True

    def can_run_in_terminal(self) -> bool:
        return True

    @classmethod
    def _get_switcher_conf_for_executable(cls, executable):
        return {"run.backend_name": cls.backend_name, f"{cls.backend_name}.executable": executable}

    def get_current_switcher_configuration(self):
        return self._get_switcher_conf_for_executable(
            get_workbench().get_option(f"{self.backend_name}.executable")
        )

    @classmethod
    def get_switcher_configuration_label(cls, conf: Dict[str, Any]) -> str:
        exe = conf[f"{cls.backend_name}.executable"]
        if is_private_python(exe) and exe == get_default_cpython_executable_for_backend():
            exe_label = tr("turcar's Python")
        else:
            exe_label = exe
        # •✶♦▸
        return cls.backend_description + "  •  " + exe_label

    @classmethod
    def get_switcher_entries(cls):
        confs = sorted(
            cls.get_last_configurations(), key=lambda conf: conf[f"{cls.backend_name}.executable"]
        )
        default_exe = get_default_cpython_executable_for_backend()
        default_conf = cls._get_switcher_conf_for_executable(default_exe)
        if default_conf not in confs:
            confs.insert(0, default_conf)

        return [
            (conf, cls.get_switcher_configuration_label(conf))
            for conf in confs
            if os.path.exists(conf[f"{cls.backend_name}.executable"])
        ]

    def open_custom_system_shell(self) -> None:
        raise NotImplementedError()

    @classmethod
    def is_valid_configuration(cls, conf: Dict[str, Any]) -> bool:
        return os.path.exists(conf[f"{cls.backend_name}.executable"])


class LocalCPythonConfigurationPage(BackendDetailsConfigPage):
    def __init__(self, master):
        super().__init__(master)

        self._configuration_variable = create_string_var(
            get_workbench().get_option("LocalCPython.executable")
        )

        entry_label = ttk.Label(self, text=tr("Python executable"))
        entry_label.grid(row=0, column=1, columnspan=2, sticky=tk.W)

        self._entry = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._configuration_variable,
            values=_get_interpreters(),
        )
        self._entry.state(["!disabled", "readonly"])

        self._entry.grid(row=1, column=1, sticky=tk.NSEW)

        self._select_button = ttk.Button(
            self,
            text="...",
            width=3,
            command=self._select_executable,
        )
        self._select_button.grid(row=1, column=2, sticky="e", padx=(10, 0))
        self.columnconfigure(1, weight=1)

        extra_text = tr("NB! turcar only supports Python %s and later") % "3.8"
        if running_on_mac_os():
            extra_text += "\n\n" + tr(
                "NB! File selection button may not work properly when selecting executables\n"
                + "from a virtual environment. In this case choose the 'activate' script instead\n"
                + "of the interpreter (or enter the path directly to the box)!"
            )
        extra_label = ttk.Label(self, text=extra_text)
        extra_label.grid(row=2, column=1, columnspan=2, pady=10, sticky="w")

        venv_text = tr(
            "You can activate an existing virtual environment also"
            + " via the right-click context menu in the file navagation"
            + " when selecting a virtual environment folder,"
            + " or the 'pyveng.cfg' file inside."
        )
        venv_text = "\n".join(textwrap.wrap(venv_text, 80))

        venv_label = ttk.Label(self, text=venv_text)
        venv_label.grid(row=3, column=1, columnspan=2, pady=10, sticky="w")

        last_row = ttk.Frame(self)
        last_row.grid(row=100, sticky="swe", column=1, columnspan=2)
        self.rowconfigure(100, weight=1)
        last_row.columnconfigure(1, weight=1)
        new_venv_link = ui_utils.create_action_label(
            last_row,
            "New virtual environment",
            self._create_venv,
        )
        new_venv_link.grid(row=0, column=1, sticky="e", pady=10)

        # self.columnconfigure(1, weight=1)

    def _select_executable(self):
        # TODO: get dir of current interpreter
        options = {"parent": self.winfo_toplevel()}
        if running_on_windows():
            options["filetypes"] = [
                (tr("Python interpreters"), "python.exe"),
                (tr("all files"), ".*"),
            ]

        filename = askopenfilename(**options)
        if not filename:
            return

        if filename.endswith("/activate"):
            filename = filename[: -len("activate")] + "python3"

        if filename:
            self._configuration_variable.set(filename)

    def _create_venv(self, event=None):
        if not _check_venv_installed(self):
            return

        messagebox.showinfo(
            "Creating new virtual environment",
            "After clicking 'OK' you need to choose an empty directory, "
            "which will be the root of your new virtual environment.",
            parent=self,
        )
        path = None
        while True:
            path = askdirectory(
                parent=self.winfo_toplevel(),
                initialdir=path,
                title=tr("Select empty directory for new virtual environment"),
            )
            if not path:
                return

            if os.listdir(path):
                messagebox.showerror(
                    tr("Bad directory"),
                    tr("Selected directory is not empty.\nSelect another or cancel."),
                    master=self,
                )
            else:
                break
        assert os.path.isdir(path)
        path = normpath_with_actual_case(path)

        proc = subprocess.Popen(
            [running.get_front_interpreter_for_subprocess(), "-m", "venv", path],
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        from turcar.workdlg import SubprocessDialog

        dlg = SubprocessDialog(self, proc, tr("Creating virtual environment"), autostart=True)
        ui_utils.show_dialog(dlg)

        if running_on_windows():
            exe_path = normpath_with_actual_case(os.path.join(path, "Scripts", "python.exe"))
        else:
            exe_path = os.path.join(path, "bin", "python3")

        if os.path.exists(exe_path):
            self._configuration_variable.set(exe_path)

    def should_restart(self):
        return self._configuration_variable.modified

    def apply(self):
        if not self.should_restart():
            return

        path = self._configuration_variable.get()
        if os.path.isfile(path):
            get_workbench().set_option("LocalCPython.executable", path)


def _get_interpreters():
    result = set()

    if running_on_windows():
        # registry
        result.update(_get_interpreters_from_windows_registry())

        for minor in [8, 9, 10, 11, 12]:
            for dir_ in [
                "C:\\Python3%d" % minor,
                "C:\\Python3%d-32" % minor,
                "C:\\Python3%d-64" % minor,
                "C:\\Program Files\\Python 3.%d" % minor,
                "C:\\Program Files\\Python 3.%d-64" % minor,
                "C:\\Program Files (x86)\\Python 3.%d" % minor,
                "C:\\Program Files (x86)\\Python 3.%d-32" % minor,
                "C:\\Program Files (x86)\\Python 3.%d-32" % minor,
                os.path.expanduser(r"~\AppData\Local\Programs\Python\Python3%d" % minor),
                os.path.expanduser(r"~\AppData\Local\Programs\Python\Python3%d-32" % minor),
            ]:
                path = os.path.join(dir_, WINDOWS_EXE)
                if os.path.exists(path):
                    result.add(normpath_with_actual_case(path))

        # other locations
        for dir_ in [
            "C:\\Anaconda3",
            "C:\\ProgramData\\Anaconda3",
            os.path.expanduser("~\\Anaconda3"),
        ]:
            path = os.path.join(dir_, WINDOWS_EXE)
            if os.path.exists(path):
                result.add(normpath_with_actual_case(path))

    else:
        # Common unix locations
        dirs = [
            "/bin",
            "/usr/bin",
            "/usr/local/bin",
            os.path.expanduser("~/.local/bin"),
            os.path.expanduser("~/anaconda3/bin"),
        ]
        for dir_ in dirs:
            # if the dir_ is just a link to another dir_, skip it
            # (not to show items twice)
            # for example on Fedora /bin -> usr/bin
            if not os.path.exists(dir_):
                continue

            apath = normpath_with_actual_case(dir_)
            if apath != dir_ and apath in dirs:
                continue
            for name in [
                "python3",
                "python3.8",
                "python3.9",
                "python3.10",
                "python3.11",
                "python3.12",
            ]:
                path = os.path.join(dir_, name)
                if os.path.exists(path):
                    result.add(path)

    if running_on_mac_os():
        for version in ["3.8", "3.9", "3.10", "3.11", "3.12"]:
            dir_ = os.path.join("/Library/Frameworks/Python.framework/Versions", version, "bin")
            path = os.path.join(dir_, "python3")

            if os.path.exists(path):
                result.add(path)

    from shutil import which

    for command in ["python3", "python3.8", "python3.9", "python3.10", "python3.11", "python3.12"]:
        path = which(command)
        if path is not None and os.path.isabs(path):
            result.add(path)

    for conf in get_workbench().get_option(f"LocalCPython.last_configurations"):
        path = conf["LocalCPython.executable"]
        if os.path.exists(path):
            result.add(normpath_with_actual_case(path))

    # TODO: remove softlinked duplicates?

    sorted_result = sorted(result)

    # bundled python
    default_path = get_default_cpython_executable_for_backend()
    if is_private_python(default_path):
        sorted_result.insert(0, default_path)

    return sorted_result


def get_default_cpython_executable_for_backend() -> str:
    if is_private_python(sys.executable) and is_virtual_executable(sys.executable):
        # Private venv. Make an exception and use base Python for default backend.
        default_path = get_base_executable()
    else:
        default_path = sys.executable.replace("pythonw.exe", "python.exe")

    # In macOS bundle the path may have ..-s
    default_path = os.path.normpath(default_path)

    """ # Too confusing:
    if running_on_mac_os():
        # try to generalize so that the path can survive Python upgrade
        ver = "{}.{}".format(*sys.version_info)
        ver_fragment = f"/Versions/{ver}/bin/"
        if ver_fragment in default_path:
            generalized = default_path.replace(ver_fragment, "/Versions/Current/bin/")
            if os.path.exists(generalized):
                return generalized
    """
    return default_path


def _get_interpreters_from_windows_registry():
    # https://github.com/python/cpython/blob/master/Tools/msi/README.txt
    # https://www.python.org/dev/peps/pep-0514/#installpath
    import winreg

    result = set()
    for key in [winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER]:
        for version in [
            "3.8",
            "3.8-32",
            "3.8-64",
            "3.9",
            "3.9-32",
            "3.9-64",
            "3.10",
            "3.10-32",
            "3.10-64",
        ]:
            for subkey in [
                "SOFTWARE\\Python\\PythonCore\\" + version + "\\InstallPath",
                "SOFTWARE\\Python\\PythonCore\\Wow6432Node\\" + version + "\\InstallPath",
            ]:
                try:
                    dir_ = winreg.QueryValue(key, subkey)
                    if dir_:
                        path = os.path.join(dir_, WINDOWS_EXE)
                        if os.path.exists(path):
                            result.add(path)
                except Exception:
                    pass

    return result


def _check_venv_installed(parent):
    try:
        import venv

        return True
    except ImportError:
        messagebox.showerror("Error", "Package 'venv' is not available.", parent=parent)
        return False



"""
TLS/SSL wrapper for socket objects.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/ssl.rst.

|see_cpython_module| :mod:`python:ssl`.

This module provides access to Transport Layer Security (previously and
widely known as “Secure Sockets Layer”) encryption and peer authentication
facilities for network sockets, both client-side and server-side.

.. admonition:: Difference to CPython
   :class: attention
The CPython version of ``ssl`` uses ``SSLError``.
      This exception does NOT exist. Instead its base class, OSError, is used.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final

from uio import StrOrBytesPath
from usocket import Socket

def wrap_socket(
    sock: Socket,
    server_side: bool = False,
    keyfile: StrOrBytesPath | None = None,
    certfile: StrOrBytesPath | None = None,
    cert_reqs: int = "CERT_NONE",
    ca_certs: str | None = None,
    do_handshake: bool = True,
    /,
) -> Socket:
    """
   Takes a `stream` *sock* (usually socket.socket instance of ``SOCK_STREAM`` type),
   and returns an instance of ssl.SSLSocket, which wraps the underlying stream in
   an SSL context. Returned object has the usual `stream` interface methods like
   ``read()``, ``write()``, etc.
   A server-side SSL socket should be created from a normal socket returned from
   :meth:`~socket.socket.accept()` on a non-SSL listening server socket.
   
   - *do_handshake* determines whether the handshake is done as part of the ``wrap_socket``
     or whether it is deferred to be done as part of the initial reads or writes
     (there is no ``do_handshake`` method as in CPython).
     For blocking sockets doing the handshake immediately is standard. For non-blocking
     sockets (i.e. when the *sock* passed into ``wrap_socket`` is in non-blocking mode)
     the handshake should generally be deferred because otherwise ``wrap_socket`` blocks
     until it completes. Note that in AXTLS the handshake can be deferred until the first
     read or write but it then blocks until completion.
   
   Depending on the underlying module implementation in a particular
   :term:`MicroPython port`, some or all keyword arguments above may be not supported.
   """

CERT_NONE: Final[int] = ...
"""
Supported values for *cert_reqs* parameter.
"""

CERT_OPTIONAL: Final[int] = ...
"""
Supported values for *cert_reqs* parameter.
"""

CERT_REQUIRED: Final[int] = ...
"""
Supported values for *cert_reqs* parameter.
"""



import os.path
import re
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Dict, List, Optional, Tuple

from pipkin.common import UserError
from pipkin.util import parse_meta_dir_name

META_ENCODING = "utf-8"
KNOWN_VID_PIDS = {(0x2E8A, 0x0005)}  # Raspberry Pi Pico

logger = getLogger(__name__)


class Adapter(ABC):
    """
    An instance of Adapter is meant to be used by single instance of Session.
    It is assumed that during the lifetime of an Adapter, sys.path stays fixed and
    distributions and sys.path directories are only manipulated via this Adapter.
    This requirement is related to the caching used in BaseAdapter.
    """

    @abstractmethod
    def get_user_packages_path(self) -> Optional[str]:
        """Unix / Windows ports return the location of user packages"""
        ...

    @abstractmethod
    def get_default_target(self) -> str:
        """Installation location if neither --user nor --target is specified"""
        ...

    @abstractmethod
    def list_dists(self, paths: List[str] = None) -> Dict[str, Tuple[str, str]]:
        """Return canonic names of the distributions mapped to their meta dir names and
        installation paths.

        If a distribution is installed to different sys.path locations, then return only the first one.
        """
        ...

    @abstractmethod
    def remove_dist(
        self, dist_name: str, target: Optional[str] = None, above_target: bool = False
    ) -> None:
        """If target is given, then remove from this directory.
        If above_path, then also remove from sys.path dirs which would hide the package at path.
        Otherwise remove the first visible instance of the dist according to sys.path.
        """
        ...

    @abstractmethod
    def read_file(self, path: str) -> bytes:
        """Path must be device's absolute path (ie. start with /)"""
        ...

    @abstractmethod
    def write_file(self, path: str, content: bytes) -> None:
        """Path must be device's absolute path (ie. start with /)"""
        ...

    @abstractmethod
    def join_path(self, *parts: str) -> str:
        ...

    @abstractmethod
    def split_dir_and_basename(self, path: str) -> Tuple[str, str]:
        ...

    @abstractmethod
    def normpath(self, path: str) -> str:
        ...

    @abstractmethod
    def get_implementation_name_and_version_prefix(self) -> Tuple[str, str]:
        ...

    @abstractmethod
    def get_mpy_cross_args(self) -> List[str]:
        ...


class DummyAdapter(Adapter):
    def get_user_packages_path(self) -> Optional[str]:
        raise NotImplementedError()

    def get_default_target(self) -> str:
        raise NotImplementedError()

    def list_dists(self, paths: List[str] = None) -> Dict[str, Tuple[str, str]]:
        raise NotImplementedError()

    def remove_dist(
        self, dist_name: str, target: Optional[str] = None, above_target: bool = False
    ) -> None:
        raise NotImplementedError()

    def read_file(self, path: str) -> bytes:
        raise NotImplementedError()

    def write_file(self, path: str, content: bytes) -> None:
        raise NotImplementedError()

    def join_path(self, *parts: str) -> str:
        raise NotImplementedError()

    def split_dir_and_basename(self, path: str) -> Tuple[str, str]:
        raise NotImplementedError()

    def normpath(self, path: str) -> str:
        raise NotImplementedError()

    def get_implementation_name_and_version_prefix(self) -> Tuple[str, str]:
        raise NotImplementedError()

    def get_mpy_cross_args(self) -> List[str]:
        raise NotImplementedError()


class BaseAdapter(Adapter, ABC):
    def __init__(self):
        self._ensured_directories = set()
        self._sys_path: Optional[List[str]] = None
        self._sys_implementation: Optional[Tuple[str, str, int]] = None

    def get_sys_path(self) -> List[str]:
        if self._sys_path is None:
            self._sys_path = self.fetch_sys_path()
        return self._sys_path

    def get_implementation_name_and_version_prefix(self) -> Tuple[str, str]:
        impl = self.get_sys_implementation()
        return impl[0], ".".join(impl[1].split(".")[:2])

    def get_mpy_cross_args(self) -> List[str]:
        impl = self.get_sys_implementation()
        sys_mpy = impl[2]
        if sys_mpy is None:
            return []

        # https://docs.micropython.org/en/latest/reference/mpyfiles.html#versioning-and-compatibility-of-mpy-files
        args = []
        arch = [
            None,
            "x86",
            "x64",
            "armv6",
            "armv6m",
            "armv7m",
            "armv7em",
            "armv7emsp",
            "armv7emdp",
            "xtensa",
            "xtensawin",
        ][sys_mpy >> 10]
        if arch:
            args.append("-march=" + arch)
        if not sys_mpy & 0x200:
            args.append("-mno-unicode")

        return args

    @abstractmethod
    def fetch_sys_path(self) -> List[str]:
        ...

    def get_sys_implementation(self) -> Tuple[str, str, int]:
        if self._sys_implementation is None:
            self._sys_implementation = self.fetch_sys_implementation()
        return self._sys_implementation

    @abstractmethod
    def fetch_sys_implementation(self) -> Tuple[str, str, int]:
        ...

    @abstractmethod
    def remove_file_if_exists(self, path: str) -> None:
        ...

    @abstractmethod
    def remove_dir_if_empty(self, path: str) -> bool:
        ...

    @abstractmethod
    def list_meta_dir_names(self, path: str, dist_name: Optional[str] = None) -> List[str]:
        """Return meta dir names from the indicated directory"""
        ...

    def get_default_target(self) -> str:
        sys_path = self.get_sys_path()
        # M5-Flow 2.0.0 has both /lib and /flash/libs
        for candidate in ["/flash/lib", "/flash/libs", "/lib"]:
            if candidate in sys_path:
                return candidate

        for entry in sys_path:
            if "lib" in entry:
                return entry
        raise AssertionError("Could not determine default target")

    def list_dists(self, paths: List[str] = None) -> Dict[str, Tuple[str, str]]:
        if not paths:
            # TODO: Consider considering only single directory
            paths = [entry for entry in self.get_sys_path() if entry.startswith("/")]

        result = {}
        for path in paths:
            for dir_name in self.list_meta_dir_names(path):
                dist_name, _ = parse_meta_dir_name(dir_name)
                if dist_name not in result:
                    result[dist_name] = dir_name, path

        return result

    def remove_dist(
        self, dist_name: str, target: Optional[str] = None, above_target: bool = False
    ) -> None:
        could_remove = False
        if target:
            result = self.check_remove_dist_from_path(dist_name, target)
            could_remove = could_remove or result
            if above_target and target in self.get_sys_path():
                for entry in self.get_sys_path():
                    if entry == "":
                        continue
                    elif entry == target:
                        break
                    else:
                        result = self.check_remove_dist_from_path(dist_name, entry)
                        could_remove = could_remove or result

        else:
            for entry in self.get_sys_path():
                if entry.startswith("/"):
                    result = self.check_remove_dist_from_path(dist_name, entry)
                    could_remove = could_remove or result
                    if result:
                        break

        if not could_remove:
            logger.warning("Could not find %r for removing", dist_name)

    def check_remove_dist_from_path(self, dist_name: str, path: str) -> bool:
        meta_dirs = self.list_meta_dir_names(path, dist_name)
        result = False
        for meta_dir_name in meta_dirs:
            self.remove_dist_by_meta_dir(path, meta_dir_name)
            result = True

        return result

    def remove_dist_by_meta_dir(self, containing_dir: str, meta_dir_name: str) -> None:
        record_bytes = self.read_file(self.join_path(containing_dir, meta_dir_name, "RECORD"))
        record_lines = record_bytes.decode(META_ENCODING).splitlines()

        package_dirs = set()
        for line in record_lines:
            rel_path, _, _ = line.split(",")
            abs_path = self.join_path(containing_dir, rel_path)
            logger.debug("Removing file %s", abs_path)
            self.remove_file_if_exists(abs_path)
            abs_dir, _ = self.split_dir_and_basename(abs_path)
            while len(abs_dir) > len(containing_dir):
                package_dirs.add(abs_dir)
                abs_dir, _ = self.split_dir_and_basename(abs_dir)

        for abs_dir in sorted(package_dirs, reverse=True):
            did_remove = self.remove_dir_if_empty(abs_dir)
            if did_remove and abs_dir in self._ensured_directories:
                self._ensured_directories.remove(abs_dir)

    def join_path(self, *parts: str) -> str:
        assert parts
        return self.get_dir_sep().join(parts)

    def split_dir_and_basename(self, path: str) -> Tuple[str, str]:
        dir_name, basename = path.rsplit(self.get_dir_sep(), maxsplit=1)
        return dir_name or None, basename or None

    def normpath(self, path: str) -> str:
        return path.replace("\\", self.get_dir_sep()).replace("/", self.get_dir_sep())

    @abstractmethod
    def get_dir_sep(self) -> str:
        ...

    def write_file(self, path: str, content: bytes) -> None:
        parent, _ = self.split_dir_and_basename(path)
        self.ensure_dir_exists(parent)
        self.write_file_in_existing_dir(path, content)

    def ensure_dir_exists(self, path: str) -> None:
        if (
            path in self._ensured_directories
            or path == "/"
            or path.endswith(":")
            or path.endswith(":\\")
        ):
            return
        else:
            parent, _ = self.split_dir_and_basename(path)
            if parent:
                self.ensure_dir_exists(parent)
            self.mkdir_in_existing_parent_exists_ok(path)
            self._ensured_directories.add(path)

    @abstractmethod
    def write_file_in_existing_dir(self, path: str, content: bytes) -> None:
        ...

    @abstractmethod
    def mkdir_in_existing_parent_exists_ok(self, path: str) -> None:
        ...


class InterpreterAdapter(BaseAdapter, ABC):
    """Base class for adapters, which communicate with an interpreter"""

    def __init__(self, executable: str):
        super().__init__()
        self._executable = executable


class ExecutableAdapter(InterpreterAdapter, ABC):
    def get_dir_sep(self) -> str:
        return os.path.sep


class LocalExecutableAdapter(ExecutableAdapter):
    ...


class SshExecutableAdapter(ExecutableAdapter):
    ...


class LocalMirrorAdapter(BaseAdapter, ABC):
    def __init__(self, base_path: str):
        super().__init__()
        self.base_path = base_path

    def get_user_packages_path(self) -> Optional[str]:
        return None

    def get_dir_sep(self) -> str:
        return "/"

    def get_mpy_cross_args(self) -> List[str]:
        return []

    def read_file(self, path: str) -> bytes:
        local_path = self.convert_to_local_path(path)
        with open(local_path, "rb") as fp:
            return fp.read()

    def write_file_in_existing_dir(self, path: str, content: bytes) -> None:
        local_path = self.convert_to_local_path(path)
        assert not os.path.isdir(local_path)

        block_size = 4 * 1024
        with open(local_path, "wb") as fp:
            while content:
                block = content[:block_size]
                content = content[block_size:]
                bytes_written = fp.write(block)
                fp.flush()
                os.fsync(fp)
                assert bytes_written == len(block)

    def remove_file_if_exists(self, path: str) -> None:
        local_path = self.convert_to_local_path(path)
        if os.path.exists(local_path):
            os.remove(local_path)

    def remove_dir_if_empty(self, path: str) -> bool:
        local_path = self.convert_to_local_path(path)
        assert os.path.isdir(local_path)
        content = os.listdir(local_path)
        if content:
            return False
        else:
            os.rmdir(local_path)
            return True

    def mkdir_in_existing_parent_exists_ok(self, path: str) -> None:
        local_path = self.convert_to_local_path(path)
        if not os.path.isdir(local_path):
            assert not os.path.exists(local_path)
            os.mkdir(local_path, 0o755)

    def convert_to_local_path(self, device_path: str) -> str:
        assert device_path.startswith("/")
        return os.path.normpath(self.base_path + device_path)

    def list_meta_dir_names(self, path: str, dist_name: Optional[str] = None) -> List[str]:
        local_path = self.convert_to_local_path(path)
        try:
            return [
                name
                for name in os.listdir(local_path)
                if name.endswith(".dist-info")
                and (dist_name is None or name.startswith(dist_name + "-"))
            ]
        except FileNotFoundError:
            # skipping non-existing dirs
            return []


class MountAdapter(LocalMirrorAdapter):
    def __init__(self, base_path: str):
        super().__init__(base_path)
        if not os.path.exists(base_path):
            raise UserError(f"Can't find mount point {base_path}")
        if os.path.isfile(base_path):
            raise UserError(f"Mount point {base_path} can't be a file")

        self._circuitpython_version = self._infer_cp_version()

    def fetch_sys_path(self) -> List[str]:
        if os.path.isdir(os.path.join(self.base_path, "lib")) or self.is_circuitpython():
            return ["", "/", ".frozen", "/lib"]
        elif os.path.isdir(os.path.join(self.base_path, "flash")):
            return ["", "/flash", "/flash/lib"]
        else:
            return ["", "/", ".frozen", "/lib"]

    def fetch_sys_implementation(self) -> Tuple[str, str, int]:
        if self._circuitpython_version:
            return ("circuitpython", self._circuitpython_version, 0)
        else:
            raise UserError("Could not determine sys.implementation")

    def is_circuitpython(self) -> bool:
        # TODO: better look into the file as well
        return os.path.isfile(os.path.join(self.base_path, "boot_out.txt"))

    def _infer_cp_version(self) -> Optional[str]:
        boot_out_path = os.path.join(self.base_path, "boot_out.txt")
        if os.path.exists(boot_out_path):
            with open(boot_out_path, encoding="utf-8") as fp:
                firmware_info = fp.readline().strip()
            match = re.match(r".*?CircuitPython (\d+\.\d+)\..+?", firmware_info)
            if match:
                return match.group(1)

        return None


class DirAdapter(LocalMirrorAdapter):
    def __init__(self, base_path: str):
        super().__init__(base_path)
        if not os.path.isdir(base_path):
            assert not os.path.exists(base_path)
            os.makedirs(base_path, mode=0o755)

    def fetch_sys_path(self) -> List[str]:
        # This means, list command without --path will consider this directory
        return ["/"]

    def fetch_sys_implementation(self) -> Tuple[str, str, int]:
        # TODO:
        return ("micropython", "1.20", 0)

    def get_default_target(self) -> str:
        return "/"


def create_adapter(port: Optional[str], mount: Optional[str], dir: Optional[str], **kw) -> Adapter:
    if port:
        from pipkin import bare_metal, serial_connection

        connection = serial_connection.SerialConnection(port)
        return bare_metal.SerialPortAdapter(connection)
    elif dir:
        return DirAdapter(dir)
    elif mount:
        return MountAdapter(mount)
    else:
        return _infer_adapter()


def _infer_adapter() -> BaseAdapter:
    from serial.tools.list_ports import comports

    candidates = [("port", p.device) for p in comports() if (p.vid, p.pid) in KNOWN_VID_PIDS]

    from .util import list_volumes

    for vol in list_volumes(skip_letters={"A"}):
        if os.path.isfile(os.path.join(vol, "boot_out.txt")):
            candidates.append(("mount", vol))

    if not candidates:
        raise UserError("Could not auto-detect target")

    if len(candidates) > 1:
        raise UserError(f"Found several possible targets: {candidates}")

    kind, arg = candidates[0]
    if kind == "port":
        from pipkin import bare_metal, serial_connection

        connection = serial_connection.SerialConnection(arg)
        return bare_metal.SerialPortAdapter(connection)
    else:
        assert kind == "mount"
        return MountAdapter(arg)



from enum import Enum


class Button(Enum):
    """
    Buttons on a brick or remote.
    """
    LEFT_DOWN = ...
    DOWN = ...
    RIGHT_DOWN = ...
    LEFT = ...
    CENTER = ...
    RIGHT = ...
    LEFT_UP = ...
    UP = ...
    BEACON = ...
    RIGHT_UP = ...



from enum import Enum


class Direction(Enum):
    """
    Rotational direction for positive speed or angle values.

    Note:
        CLOCKWISE: A positive speed value should make the motor move clockwise.
        COUNTERCLOCKWISE: A positive speed value should make the motor move counterclockwise.

        By default, the positive direction is set as clockwise.

    """
    CLOCKWISE = "CLOCKWISE"
    COUNTERCLOCKWISE = "COUNTERCLOCKWISE"



"""
random numbers.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/random.rst.
========================================

.. module:: random
   :synopsis: random numbers

This module implements a pseudo-random number generator (PRNG).

|see_cpython_module| :mod:`python:random` .

.. note::

   The following notation is used for intervals:

   - () are open interval brackets and do not include their endpoints.
     For example, (0, 1) means greater than 0 and less than 1.
     In set notation: (0, 1) = {x | 0 < x < 1}.

   - [] are closed interval brackets which include all their limit points.
     For example, [0, 1] means greater than or equal to 0 and less than
     or equal to 1.
     In set notation: [0, 1] = {x | 0 <= x <= 1}.

.. note::

   The :func:`randrange`, :func:`randint` and :func:`choice` functions are only
   available if the ``MICROPY_PY_URANDOM_EXTRA_FUNCS`` configuration option is
   enabled.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import TypeVar, runtime_checkable, Protocol, overload

_T = TypeVar("_T")
@runtime_checkable
class Subscriptable(Protocol[_T]):
    """A `Protocol` (structurally typed) for an object that is subscriptable and of finite length."""

    __slots__ = ()
    def __len__(self) -> int:
        """Number of elements, normally called via `len(x)` where `x` is an object that implements this protocol."""
    def __getitem__(self, index: int) -> _T:
        """
        Element at the given index, 
        normally called via `x[index]` where `x` is an object that implements this protocol.
        """

def getrandbits(n: int, /) -> int:
    """
    Return an integer with *n* random bits (0 <= n <= 32).
   """

def randint(a: int, b: int, /) -> int:
    """
    Return a random integer in the range [*a*, *b*].
   """

@overload
def randrange(stop: int, /) -> int:
    """
    The first form returns a random integer from the range [0, *stop*).
    The second form returns a random integer from the range [*start*, *stop*).
    The third form returns a random integer from the range [*start*, *stop*) in
    steps of *step*.  For instance, calling ``randrange(1, 10, 2)`` will
    return odd numbers between 1 and 9 inclusive.
   """

@overload
def randrange(start: int, stop: int, /) -> int:
    """
    The first form returns a random integer from the range [0, *stop*).
    The second form returns a random integer from the range [*start*, *stop*).
    The third form returns a random integer from the range [*start*, *stop*) in
    steps of *step*.  For instance, calling ``randrange(1, 10, 2)`` will
    return odd numbers between 1 and 9 inclusive.
   """

@overload
def randrange(start: int, stop: int, step: int, /) -> int:
    """
    The first form returns a random integer from the range [0, *stop*).
    The second form returns a random integer from the range [*start*, *stop*).
    The third form returns a random integer from the range [*start*, *stop*) in
    steps of *step*.  For instance, calling ``randrange(1, 10, 2)`` will
    return odd numbers between 1 and 9 inclusive.
   """

def random() -> float:
    """
    Return a random floating point number in the range [0.0, 1.0).
   """

def uniform(a: float, b: float) -> float:
    """
    Return a random floating point number N such that *a* <= N <= *b* for *a* <= *b*,
    and *b* <= N <= *a* for *b* < *a*.
   """

def seed(n: int | None = None, /) -> None:
    """
    Initialise the random number generator module with the seed *n* which should
    be an integer.  When no argument (or ``None``) is passed in it will (if
    supported by the port) initialise the PRNG with a true random number
    (usually a hardware generated random number).
    
    The ``None`` case only works if ``MICROPY_PY_URANDOM_SEED_INIT_FUNC`` is
    enabled by the port, otherwise it raises ``ValueError``.
   """

def choice(sequence: Subscriptable, /) -> None:
    """
    Chooses and returns one item at random from *sequence* (tuple, list or
    any object that supports the subscript operation).
   """
